"""Pipeline snapshots [8ad841ad9bfe].

Revision ID: 8ad841ad9bfe
Revises: a71cca37e42d
Create Date: 2025-07-22 11:20:28.777003

"""

import uuid

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.dialects import mysql

from zenml.utils.time_utils import utc_now

# revision identifiers, used by Alembic.
revision = "8ad841ad9bfe"
down_revision = "a71cca37e42d"
branch_labels = None
depends_on = None


def rename_pipeline_deployment_to_pipeline_snapshot() -> None:
    """Rename the pipeline deployment table to pipeline snapshot."""
    # 1. Drop existing foreign keys referencing the pipeline deployment table
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_pipeline_run_deployment_id_pipeline_deployment",
            type_="foreignkey",
        )

    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_run_template_source_deployment_id_pipeline_deployment",
            type_="foreignkey",
        )

    with op.batch_alter_table("step_configuration", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_step_configuration_deployment_id_pipeline_deployment",
            type_="foreignkey",
        )

    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_step_run_deployment_id_pipeline_deployment", type_="foreignkey"
        )

    # 2. Drop foreign keys of the pipeline deployment tables themselves, which
    # we will have to rename
    with op.batch_alter_table("pipeline_deployment", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_pipeline_deployment_build_id_pipeline_build",
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_code_reference_id_code_reference",
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_pipeline_id_pipeline", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_project_id_project", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_schedule_id_schedule", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_stack_id_stack", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_pipeline_deployment_user_id_user", type_="foreignkey"
        )

    # 3. Rename table
    op.rename_table("pipeline_deployment", "pipeline_snapshot")

    # 4. Recreate foreign keys on the renamed table
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_build_id_pipeline_build",
            "pipeline_build",
            ["build_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_code_reference_id_code_reference",
            "code_reference",
            ["code_reference_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_pipeline_id_pipeline",
            "pipeline",
            ["pipeline_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_project_id_project",
            "project",
            ["project_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_schedule_id_schedule",
            "schedule",
            ["schedule_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_stack_id_stack",
            "stack",
            ["stack_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_user_id_user",
            "user",
            ["user_id"],
            ["id"],
            ondelete="SET NULL",
        )

    # 5. Drop more constraints that need to be renamed, and alter the column
    # names. For some reason, we can't use the same `batch_alter_table` to
    # rename columns and then create new constraints, so we have to do it in
    # separate calls.
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_orchestrator_run_id_for_deployment_id", type_="unique"
        )

        batch_op.alter_column(
            "deployment_id",
            existing_type=sa.CHAR(length=32),
            new_column_name="snapshot_id",
        )

    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.alter_column(
            "source_deployment_id",
            existing_type=sa.CHAR(length=32),
            new_column_name="source_snapshot_id",
        )

    with op.batch_alter_table("step_configuration", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_step_name_for_deployment", type_="unique"
        )

        batch_op.alter_column(
            "deployment_id",
            existing_type=sa.CHAR(length=32),
            new_column_name="snapshot_id",
        )

    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.alter_column(
            "deployment_id",
            existing_type=sa.CHAR(length=32),
            new_column_name="snapshot_id",
        )

    # 6. Add back the constraints with correct columns and names
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_orchestrator_run_id_for_snapshot_id",
            ["snapshot_id", "orchestrator_run_id"],
        )
        batch_op.create_foreign_key(
            "fk_pipeline_run_snapshot_id_pipeline_snapshot",
            "pipeline_snapshot",
            ["snapshot_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("run_template", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_run_template_source_snapshot_id_pipeline_snapshot",
            "pipeline_snapshot",
            ["source_snapshot_id"],
            ["id"],
            ondelete="SET NULL",
        )
    with op.batch_alter_table("step_configuration", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "unique_step_name_for_snapshot", ["snapshot_id", "name"]
        )
        batch_op.create_foreign_key(
            "fk_step_configuration_snapshot_id_pipeline_snapshot",
            "pipeline_snapshot",
            ["snapshot_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_step_run_snapshot_id_pipeline_snapshot",
            "pipeline_snapshot",
            ["snapshot_id"],
            ["id"],
            ondelete="CASCADE",
        )


def add_unlisted_pipeline_if_necessary() -> None:
    """Create pipelines for orphaned snapshots without a pipeline reference."""
    connection = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(
        bind=connection, only=("pipeline_snapshot", "pipeline", "pipeline_run")
    )

    pipeline_snapshot_table = sa.Table("pipeline_snapshot", meta)
    pipeline_table = sa.Table("pipeline", meta)
    pipeline_run_table = sa.Table("pipeline_run", meta)

    projects_with_orphaned_snapshots = connection.execute(
        sa.select(
            pipeline_snapshot_table.c.project_id,
        )
        .where(pipeline_snapshot_table.c.pipeline_id.is_(None))
        .group_by(pipeline_snapshot_table.c.project_id)
    ).fetchall()

    if not projects_with_orphaned_snapshots:
        return

    now = utc_now()

    for project_id_row in projects_with_orphaned_snapshots:
        existing_pipeline = connection.execute(
            sa.select(pipeline_table.c.id)
            .where(pipeline_table.c.project_id == project_id_row.project_id)
            .where(pipeline_table.c.name == "unlisted")
        ).fetchone()

        if existing_pipeline:
            unlisted_pipeline_id = existing_pipeline[0]
        else:
            unlisted_pipeline_id = str(uuid.uuid4()).replace("-", "")
            connection.execute(
                sa.insert(pipeline_table).values(
                    id=unlisted_pipeline_id,
                    created=now,
                    updated=now,
                    name="unlisted",
                    description="Auto-created pipeline for snapshots without "
                    "pipeline reference",
                    project_id=project_id_row.project_id,
                    user_id=None,
                )
            )

        connection.execute(
            sa.update(pipeline_snapshot_table)
            .where(
                pipeline_snapshot_table.c.project_id
                == project_id_row.project_id
            )
            .where(pipeline_snapshot_table.c.pipeline_id.is_(None))
            .values(pipeline_id=unlisted_pipeline_id)
        )

        connection.execute(
            sa.update(pipeline_run_table)
            .where(
                pipeline_run_table.c.project_id == project_id_row.project_id
            )
            .where(pipeline_run_table.c.pipeline_id.is_(None))
            .values(pipeline_id=unlisted_pipeline_id)
        )


def migrate_run_templates() -> None:
    """Migrate run templates into the snapshot table."""
    connection = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(
        bind=connection,
        only=("pipeline_snapshot", "run_template", "tag_resource"),
    )

    pipeline_snapshot_table = sa.Table("pipeline_snapshot", meta)
    run_template_table = sa.Table("run_template", meta)
    tag_resource_table = sa.Table("tag_resource", meta)

    # Step 1: Update the `template_id` of snapshots to the source snapshot
    # id
    snapshot_template_mapping = connection.execute(
        sa.select(
            pipeline_snapshot_table.c.id,
            run_template_table.c.source_snapshot_id,
        )
        .outerjoin(
            run_template_table,
            run_template_table.c.id == pipeline_snapshot_table.c.template_id,
        )
        .where(pipeline_snapshot_table.c.template_id.is_not(None))
    ).fetchall()

    snapshot_updates = [
        {"id_": snapshot_id, "source_snapshot_id": source_snapshot_id}
        for snapshot_id, source_snapshot_id in snapshot_template_mapping
    ]
    if snapshot_updates:
        connection.execute(
            sa.update(pipeline_snapshot_table)
            .where(pipeline_snapshot_table.c.id == sa.bindparam("id_"))
            .values(source_snapshot_id=sa.bindparam("source_snapshot_id")),
            snapshot_updates,
        )

    # Step 2: Migrate tags from run templates to their source snapshots
    tag_run_template_mapping = connection.execute(
        sa.select(
            tag_resource_table.c.tag_id,
            run_template_table.c.source_snapshot_id,
        )
        .join(
            run_template_table,
            run_template_table.c.id == tag_resource_table.c.resource_id,
        )
        .where(tag_resource_table.c.resource_type == "run_template")
    ).fetchall()

    now = utc_now()

    tag_resource_insertions = [
        {
            "id": str(uuid.uuid4()).replace("-", ""),
            "created": now,
            "updated": now,
            "tag_id": tag_id,
            "resource_id": source_snapshot_id,
            "resource_type": "pipeline_snapshot",
        }
        for tag_id, source_snapshot_id in tag_run_template_mapping
    ]
    if tag_resource_insertions:
        op.bulk_insert(tag_resource_table, tag_resource_insertions)

    # Step 3: Migrate non-hidden run templates to their source snapshots
    # If there are multiple templates for the same source snapshot, use the
    # name/description of the latest one.
    latest_templates_subquery = (
        sa.select(
            run_template_table.c.source_snapshot_id,
            sa.func.max(run_template_table.c.created).label("created"),
        )
        .where(run_template_table.c.hidden.is_(False))
        .where(run_template_table.c.source_snapshot_id.is_not(None))
        .group_by(run_template_table.c.source_snapshot_id)
        .subquery()
    )
    latest_templates_query = sa.select(
        run_template_table.c.name,
        run_template_table.c.description,
        run_template_table.c.source_snapshot_id,
    ).join(
        latest_templates_subquery,
        sa.and_(
            latest_templates_subquery.c.source_snapshot_id
            == run_template_table.c.source_snapshot_id,
            run_template_table.c.created
            == latest_templates_subquery.c.created,
        ),
    )

    snapshot_updates = [
        {
            "id_": source_snapshot_id,
            "name": template_name,
            "description": template_description,
        }
        for template_name, template_description, source_snapshot_id in connection.execute(
            latest_templates_query
        ).fetchall()
    ]
    if snapshot_updates:
        connection.execute(
            sa.update(pipeline_snapshot_table)
            .where(pipeline_snapshot_table.c.id == sa.bindparam("id_"))
            .values(
                name=sa.bindparam("name"),
                description=sa.bindparam("description"),
            ),
            snapshot_updates,
        )


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    rename_pipeline_deployment_to_pipeline_snapshot()
    add_unlisted_pipeline_if_necessary()

    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "name", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "description",
                sa.String(length=16777215).with_variant(
                    mysql.MEDIUMTEXT(), "mysql"
                ),
                nullable=True,
            )
        )
        batch_op.drop_constraint(
            "fk_pipeline_snapshot_pipeline_id_pipeline", type_="foreignkey"
        )
        batch_op.alter_column(
            "pipeline_id", existing_type=sa.CHAR(length=32), nullable=False
        )
        batch_op.create_unique_constraint(
            "unique_name_for_pipeline_id", ["pipeline_id", "name"]
        )
        batch_op.create_foreign_key(
            "fk_pipeline_snapshot_pipeline_id_pipeline",
            "pipeline",
            ["pipeline_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.add_column(
            sa.Column(
                "source_snapshot_id",
                sqlmodel.sql.sqltypes.GUID(),
                nullable=True,
            )
        )

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "triggered_by", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "triggered_by_type",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )

    migrate_run_templates()


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: Downgrade not implemented.
    """
    raise NotImplementedError("Downgrade not implemented")



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utilities for caching."""

import hashlib
import os
from typing import TYPE_CHECKING, List, Mapping, Optional
from uuid import UUID

from zenml.client import Client
from zenml.constants import CODE_HASH_PARAMETER_NAME
from zenml.enums import ExecutionStatus, SorterOps
from zenml.logger import get_logger
from zenml.orchestrators import step_run_utils
from zenml.utils import source_code_utils, source_utils

if TYPE_CHECKING:
    from zenml.artifact_stores import BaseArtifactStore
    from zenml.config.step_configurations import Step
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotResponse,
        StepRunResponse,
    )
    from zenml.models.v2.core.step_run import (
        StepRunInputResponse,
    )
    from zenml.stack import Stack


logger = get_logger(__name__)


def generate_cache_key(
    step: "Step",
    input_artifacts: Mapping[str, List["StepRunInputResponse"]],
    artifact_store: "BaseArtifactStore",
    project_id: "UUID",
) -> str:
    """Generates a cache key for a step run.

    If a step has caching enabled, the cache key will be used to find existing
    equivalent step runs.

    The cache key is a MD5 hash of:
    - the project ID,
    - the artifact store ID and path,
    - the source code that defines the step,
    - the parameters of the step,
    - the names and IDs of the input artifacts of the step,
    - the names and source codes of the output artifacts of the step,
    - the source codes of the output materializers of the step.
    - additional custom caching parameters of the step.
    - the environment variables defined for the step.
    - the secrets defined for the step.
    - other elements of the cache policy defined for the step.

    Args:
        step: The step to generate the cache key for.
        input_artifacts: The input artifacts for the step.
        artifact_store: The artifact store of the active stack.
        project_id: The ID of the active project.

    Raises:
        ValueError: If some file dependencies are outside the source root or
            missing.
        ValueError: If the cache function is invalid.
        RuntimeError: If executing the cache function failed.

    Returns:
        A cache key.
    """
    cache_policy = step.config.cache_policy
    hash_ = hashlib.md5()  # nosec

    # Project ID
    hash_.update(project_id.bytes)

    # Artifact store ID and path
    hash_.update(artifact_store.id.bytes)
    hash_.update(artifact_store.path.encode())

    if artifact_store.custom_cache_key:
        hash_.update(artifact_store.custom_cache_key)

    # Step source. This currently only uses the string representation of the
    # source (e.g. my_module.step_class) instead of the full source to keep
    # the caching behavior of previous versions and to not invalidate caching
    # when committing some unrelated files
    hash_.update(step.spec.source.import_path.encode())

    if cache_policy.include_step_parameters:
        for key, value in sorted(step.config.parameters.items()):
            hash_.update(key.encode())
            hash_.update(str(value).encode())

    # Input artifacts
    for name, artifact_versions in input_artifacts.items():
        if name in (cache_policy.ignored_inputs or []):
            continue

        hash_.update(name.encode())

        for artifact_version in artifact_versions:
            if (
                artifact_version.content_hash
                and cache_policy.include_artifact_values
            ):
                hash_.update(artifact_version.content_hash.encode())
            elif cache_policy.include_artifact_ids:
                hash_.update(artifact_version.id.bytes)

            if artifact_version.chunk_index is not None:
                hash_.update(str(artifact_version.chunk_index).encode())
                chunk_size = artifact_version.chunk_size or 1
                hash_.update(str(chunk_size).encode())

    # Output artifacts and materializers
    for name, output in step.config.outputs.items():
        hash_.update(name.encode())
        for source in output.materializer_source:
            hash_.update(source.import_path.encode())

    # Custom caching parameters
    for key, value in sorted(step.config.caching_parameters.items()):
        if (
            key == CODE_HASH_PARAMETER_NAME
            and not cache_policy.include_step_code
        ):
            continue

        hash_.update(key.encode())
        hash_.update(str(value).encode())

    # User-defined environment variables
    for key, value in sorted(step.config.environment.items()):
        hash_.update(key.encode())
        hash_.update(str(value).encode())

    # User-defined secrets
    for secret_name_or_id in sorted([str(s) for s in step.config.secrets]):
        hash_.update(secret_name_or_id.encode())

    if file_dependencies := cache_policy.file_dependencies:
        source_root = source_utils.get_source_root()

        absolute_paths = []
        for file_path in file_dependencies:
            if os.path.isabs(file_path):
                if not os.path.relpath(file_path, source_root):
                    raise ValueError(
                        f"Cache policy file dependency `{file_path}` is not "
                        f"within your source root `{source_root}`. Only files "
                        "within your source root are allowed."
                    )
            else:
                file_path = os.path.abspath(
                    os.path.join(source_root, file_path)
                )

            if not os.path.exists(file_path):
                raise ValueError(
                    f"Missing file for cache computation: `{file_path}`. "
                    "Please make sure to specify values in "
                    "`CachePolicy.file_dependencies` as a path relative to "
                    f"your source root ({source_root})."
                )

            absolute_paths.append(file_path)

        for file_path in sorted(absolute_paths):
            # Use relative path to source root as key so we don't generate
            # different hashes on different machines
            relative_path = os.path.relpath(file_path, source_root)
            hash_.update(relative_path.encode())
            with open(file_path, "rb") as f:
                while chunk := f.read(1024 * 1024):
                    hash_.update(chunk)

    if cache_policy.source_dependencies:
        source_dependencies = [
            source.import_path for source in cache_policy.source_dependencies
        ]
        for source_path in sorted(source_dependencies):
            hash_.update(source_path.encode())
            object_ = source_utils.load(source_path)
            source_code = source_code_utils.get_source_code(object_)
            hash_.update(source_code.encode())

    if cache_policy.cache_func:
        cache_func = source_utils.load(cache_policy.cache_func)

        if not callable(cache_func):
            raise ValueError(
                f"The cache function {cache_policy.cache_func.import_path} is "
                "not callable."
            )

        try:
            result = cache_func()
        except Exception as e:
            raise RuntimeError("Failed to run cache function.") from e

        if not isinstance(result, str):
            raise ValueError(
                f"The cache function {cache_policy.cache_func.import_path} "
                "must return a string."
            )

        hash_.update(result.encode())

    return hash_.hexdigest()


def get_cached_step_run(cache_key: str) -> Optional["StepRunResponse"]:
    """If a given step can be cached, get the corresponding existing step run.

    A step run can be cached if there is an existing step run in the same
    project which has the same cache key and was successfully executed.

    Args:
        cache_key: The cache key of the step.

    Returns:
        The existing step run if the step can be cached, otherwise None.
    """
    client = Client()

    cache_candidates = client.list_run_steps(
        project=client.active_project.id,
        cache_key=cache_key,
        cache_expired=False,
        status=ExecutionStatus.COMPLETED,
        sort_by=f"{SorterOps.DESCENDING}:created",
        size=1,
        hydrate=True,
    ).items

    if cache_candidates:
        return cache_candidates[0]
    return None


def create_cached_step_runs_and_prune_snapshot(
    snapshot: "PipelineSnapshotResponse",
    pipeline_run: "PipelineRunResponse",
    stack: "Stack",
) -> bool:
    """Create cached step runs and prune the cached steps from the snapshot.

    Args:
        snapshot: The pipeline snapshot.
        pipeline_run: The pipeline run for which to create the step runs.
        stack: The stack on which the pipeline run is happening.

    Returns:
        Whether an actual pipeline run is still required.
    """
    cached_invocations = step_run_utils.create_cached_step_runs(
        snapshot=snapshot,
        pipeline_run=pipeline_run,
        stack=stack,
    )

    for invocation_id in cached_invocations:
        # Remove the cached step invocations from the snapshot so
        # the orchestrator does not try to run them
        snapshot.step_configurations.pop(invocation_id)

    for step in snapshot.step_configurations.values():
        for invocation_id in cached_invocations:
            if invocation_id in step.spec.upstream_steps:
                step.spec.upstream_steps.remove(invocation_id)

    if len(snapshot.step_configurations) == 0:
        # All steps were cached, we update the pipeline run status and
        # don't actually use the orchestrator to run the pipeline
        logger.info("All steps of the pipeline run were cached.")
        return False

    return True



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization of the MLX integration."""

import platform
import sys
from typing import List, Optional

from zenml.integrations.constants import MLX
from zenml.integrations.integration import Integration

def _is_supported_platform() -> bool:
    """Check if the current platform supports MLX.
    
    MLX requires:
    - macOS with ARM64 (Apple Silicon)
    - Linux (any architecture)
    
    Returns:
        True if platform is supported, False otherwise.
    """
    if sys.platform == "linux":
        return True
    elif sys.platform == "darwin":
        # MLX only supports Apple Silicon Macs, not Intel
        machine = platform.machine().lower()
        return machine in ("arm64", "aarch64")
    return False


class MLXIntegration(Integration):
    """Definition of MLX array integration for ZenML."""

    NAME = MLX

    @classmethod
    def check_installation(cls) -> bool:
        return False if not _is_supported_platform() else super().check_installation()

    @classmethod
    def get_requirements(
        cls,
        target_os: Optional[str] = None,
        python_version: Optional[str] = None,
    ) -> List[str]:
        # sys.platform is "darwin", while platform.system() is "Darwin",
        # similarly on Linux.
        target_os = (target_os or sys.platform).lower()
        if target_os == "darwin":
            # Only return requirements if on Apple Silicon
            machine = platform.machine().lower()
            return ["mlx"] if machine in ("arm64", "aarch64") else []
        elif target_os == "linux":
            return ["mlx[cpu]"]
        else:
            return []

    @classmethod
    def activate(cls) -> None:
        """Activates the integration."""
        from zenml.integrations.mlx import materializer  # noqa



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Materializer for JSON-serializable dataclasses."""

import hashlib
import inspect
import json
import os
from dataclasses import is_dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Tuple, Type

from pydantic import TypeAdapter

from zenml.enums import ArtifactType, VisualizationType
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import yaml_utils

if TYPE_CHECKING:
    from zenml.metadata.metadata_types import MetadataType

DEFAULT_FILENAME = "data.json"


def _is_json_serializable_dataclass_type(data_type: Type[Any]) -> bool:
    """Checks whether a type is a JSON-serializable dataclass type.

    Args:
        data_type: The type to check.

    Returns:
        Whether the type is a dataclass type that Pydantic can serialize
        to JSON.
    """
    if not inspect.isclass(data_type) or not is_dataclass(data_type):
        return False

    try:
        TypeAdapter(data_type).json_schema(mode="serialization")
    except Exception:
        return False

    return True


class DataclassMaterializer(BaseMaterializer):
    """Dataclass materializer."""

    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA
    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (object,)
    # Dataclasses do not share a common base class, so we cannot auto-register
    # it in the registry. Instead, it is manually handled in some special logic
    # in the materializer registry.
    SKIP_REGISTRATION: ClassVar[bool] = True

    @classmethod
    def can_save_type(cls, data_type: Type[Any]) -> bool:
        """Checks whether the materializer can save a data type.

        Args:
            data_type: The type to check.

        Returns:
            Whether the type can be saved.
        """
        return _is_json_serializable_dataclass_type(data_type)

    @classmethod
    def can_load_type(cls, data_type: Type[Any]) -> bool:
        """Checks whether the materializer can load a data type.

        Args:
            data_type: The type to check.

        Returns:
            Whether the type can be loaded.
        """
        return _is_json_serializable_dataclass_type(data_type)

    def load(self, data_type: Type[Any]) -> Any:
        """Reads a dataclass instance from JSON.

        Args:
            data_type: The type of the data to read.

        Returns:
            The deserialized dataclass instance.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        contents = yaml_utils.read_json(data_path)
        return TypeAdapter(data_type).validate_python(contents)

    def save(self, data: Any) -> None:
        """Serializes a dataclass instance to JSON.

        Args:
            data: The dataclass instance to store.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        adapter = TypeAdapter(type(data))
        yaml_utils.write_json(
            data_path,
            adapter.dump_python(data, mode="json"),
        )

    def extract_metadata(self, data: Any) -> Dict[str, "MetadataType"]:
        """Extracts metadata from the given dataclass instance.

        Args:
            data: The dataclass instance to inspect.

        Returns:
            The extracted metadata.
        """
        return {"schema": TypeAdapter(type(data)).json_schema()}

    def compute_content_hash(self, data: Any) -> Optional[str]:
        """Computes a deterministic content hash for the dataclass.

        Args:
            data: The dataclass instance to hash.

        Returns:
            The content hash.
        """
        adapter = TypeAdapter(type(data))
        data = adapter.dump_python(data, mode="json")
        hash_ = hashlib.md5(usedforsecurity=False)
        hash_.update(json.dumps(data, sort_keys=True).encode())
        return hash_.hexdigest()

    def save_visualizations(self, data: Any) -> Dict[str, "VisualizationType"]:
        """Saves JSON visualizations for the given dataclass instance.

        Args:
            data: The dataclass instance to visualize.

        Returns:
            A dictionary of visualization URIs and their types.
        """
        data_path = os.path.join(self.uri, DEFAULT_FILENAME)
        return {data_path.replace("\\", "/"): VisualizationType.JSON}



"""Add concurrency trigger column [b949c5d1bc0d].

Revision ID: b949c5d1bc0d
Revises: 3855b5d051cd
Create Date: 2026-02-18 10:46:09.881375

"""

import sqlalchemy as sa
from alembic import op

from zenml.enums import TriggerRunConcurrency

# revision identifiers, used by Alembic.
revision = "b949c5d1bc0d"
down_revision = "3855b5d051cd"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("concurrency", sa.VARCHAR(length=20), nullable=True)
        )

    connection = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(bind=connection, only=["trigger"])
    trigger_table = sa.Table("trigger", meta)
    connection.execute(
        sa.update(trigger_table).values(
            concurrency=TriggerRunConcurrency.SKIP.value
        )
    )

    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.alter_column(
            "concurrency", existing_type=sa.VARCHAR(length=20), nullable=False
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("trigger", schema=None) as batch_op:
        batch_op.drop_column("concurrency")



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization for ZenML."""

from importlib.metadata import entry_points
import os
from typing import Any

def _run_init_hooks() -> None:
    """Runs installed ZenML init hooks."""
    for entrypoint in entry_points().select(group="zenml.init_hooks"):
        entrypoint.load()()

_run_init_hooks()

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# Set the version
with open(os.path.join(ROOT_DIR, "VERSION")) as version_file:
    __version__: str = version_file.read().strip()

# Initialize logging
from zenml.logger import init_logging  # noqa

init_logging()

def __getattr__(name: str) -> Any:
    # We allow directly accessing the entrypoint module as `zenml.entrypoint`
    # as this is needed for some orchestrators. Instead of directly importing
    # the entrypoint module here, we import it dynamically. This avoids a
    # warning when running the `zenml.entrypoints.entrypoint` module directly.
    if name == "entrypoint":
        from zenml.entrypoints import entrypoint
        return entrypoint

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

# Need to import zenml.models before zenml.config to avoid circular imports
from zenml.models import *  # noqa: F401

# Define public Python API
from zenml.utils.dashboard_utils import show_dashboard as show
from zenml.artifacts.utils import (
    log_artifact_metadata,
    save_artifact,
    load_artifact,
    register_artifact,
)
from zenml.model.utils import (
    log_model_metadata,
    link_artifact_to_model,
)
from zenml.artifacts.artifact_config import ArtifactConfig
from zenml.artifacts.external_artifact import ExternalArtifact
from zenml.model.model import Model
from zenml.pipelines import get_pipeline_context, pipeline
from zenml.steps import step, get_step_context
from zenml.steps.utils import log_step_metadata
from zenml.utils.metadata_utils import log_metadata, bulk_log_metadata
from zenml.utils.tag_utils import Tag, add_tags, remove_tags
from zenml.execution.pipeline.dynamic.utils import unmapped, wait

__all__ = [
    "add_tags",
    "remove_tags",
    "Tag",
    "ArtifactConfig",
    "ExternalArtifact",
    "get_pipeline_context",
    "get_step_context",
    "load_artifact",
    "log_metadata",
    "bulk_log_metadata",
    "log_artifact_metadata",
    "log_model_metadata",
    "log_step_metadata",
    "Model",
    "link_artifact_to_model",
    "pipeline",
    "save_artifact",
    "register_artifact",
    "show",
    "step",
    "unmapped",
    "wait",
]



"""Add model version producer run unique constraint [a1237ba94fd8].

Revision ID: a1237ba94fd8
Revises: 26351d482b9e
Create Date: 2024-12-13 10:28:55.432414

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "a1237ba94fd8"
down_revision = "26351d482b9e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "producer_run_id_if_numeric",
                sqlmodel.sql.sqltypes.GUID(),
                nullable=True,
            )
        )

    # Set the producer_run_id_if_numeric column to the model version ID for
    # existing rows
    connection = op.get_bind()
    metadata = sa.MetaData()
    metadata.reflect(only=("model_version",), bind=connection)
    model_version_table = sa.Table("model_version", metadata)

    connection.execute(
        model_version_table.update().values(
            producer_run_id_if_numeric=model_version_table.c.id
        )
    )

    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.alter_column(
            "producer_run_id_if_numeric",
            existing_type=sqlmodel.sql.sqltypes.GUID(),
            nullable=False,
        )
        batch_op.create_unique_constraint(
            "unique_numeric_version_for_pipeline_run",
            ["model_id", "producer_run_id_if_numeric"],
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.drop_constraint(
            "unique_numeric_version_for_pipeline_run", type_="unique"
        )
        batch_op.drop_column("producer_run_id_if_numeric")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for topological sort.

Implementation heavily inspired by TFX:
https://github.com/tensorflow/tfx/blob/master/tfx/utils/topsort.py
"""

from typing import Callable, List, Sequence, TypeVar

from zenml.logger import get_logger

logger = get_logger(__name__)

NodeT = TypeVar("NodeT")


def topsorted_layers(
    nodes: Sequence[NodeT],
    get_node_id_fn: Callable[[NodeT], str],
    get_parent_nodes: Callable[[NodeT], List[NodeT]],
    get_child_nodes: Callable[[NodeT], List[NodeT]],
) -> List[List[NodeT]]:
    """Sorts the DAG of nodes in topological order.

    Args:
        nodes: A sequence of nodes.
        get_node_id_fn: Callable that returns a unique text identifier for a node.
        get_parent_nodes: Callable that returns a list of parent nodes for a node.
            If a parent node's id is not found in the list of node ids, that parent
            node will be omitted.
        get_child_nodes: Callable that returns a list of child nodes for a node.
            If a child node's id is not found in the list of node ids, that child
            node will be omitted.

    Returns:
        A list of topologically ordered node layers. Each layer of nodes is sorted
        by its node id given by `get_node_id_fn`.

    Raises:
        RuntimeError: If the input nodes don't form a DAG.
        ValueError: If the nodes are not unique.
    """
    # Make sure the nodes are unique.
    node_ids = set(get_node_id_fn(n) for n in nodes)
    if len(node_ids) != len(nodes):
        raise ValueError("Nodes must have unique ids.")

    # The outputs of get_(parent|child)_nodes should always be deduplicated,
    # and references to unknown nodes should be removed.
    def _apply_and_clean(
        func: Callable[[NodeT], List[NodeT]], func_name: str, node: NodeT
    ) -> List[NodeT]:
        seen_inner_node_ids = set()
        result = []
        for inner_node in func(node):
            inner_node_id = get_node_id_fn(inner_node)
            if inner_node_id in seen_inner_node_ids:
                logger.warning(
                    "Duplicate node_id %s found when calling %s on node %s. "
                    "This entry will be ignored.",
                    inner_node_id,
                    func_name,
                    node,
                )
            elif inner_node_id not in node_ids:
                logger.warning(
                    "node_id %s found when calling %s on node %s, but this node_id is "
                    "not found in the set of input nodes %s. This entry will be "
                    "ignored.",
                    inner_node_id,
                    func_name,
                    node,
                    node_ids,
                )
            else:
                seen_inner_node_ids.add(inner_node_id)
                result.append(inner_node)

        return result

    def get_clean_parent_nodes(node: NodeT) -> List[NodeT]:
        return _apply_and_clean(get_parent_nodes, "get_parent_nodes", node)

    def get_clean_child_nodes(node: NodeT) -> List[NodeT]:
        return _apply_and_clean(get_child_nodes, "get_child_nodes", node)

    # The first layer contains nodes with no incoming edges.
    layer = [node for node in nodes if not get_clean_parent_nodes(node)]

    visited_node_ids = set()
    layers = []
    while layer:
        layer = sorted(layer, key=get_node_id_fn)
        layers.append(layer)

        next_layer = []
        for node in layer:
            visited_node_ids.add(get_node_id_fn(node))
            for child_node in get_clean_child_nodes(node):
                # Include the child node if all its parents are visited. If the child
                # node is part of a cycle, it will never be included since it will have
                # at least one unvisited parent node which is also part of the cycle.
                parent_node_ids = set(
                    get_node_id_fn(p)
                    for p in get_clean_parent_nodes(child_node)
                )
                if parent_node_ids.issubset(visited_node_ids):
                    next_layer.append(child_node)
        layer = next_layer

    num_output_nodes = sum(len(layer) for layer in layers)
    # Nodes in cycles are not included in layers; raise an error if this happens.
    if num_output_nodes < len(nodes):
        raise RuntimeError("Cannot sort graph because it contains a cycle.")
    # This should never happen; raise an error if this occurs.
    if num_output_nodes > len(nodes):
        raise RuntimeError("Unknown error occurred while sorting DAG.")

    return layers



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Sagemaker Step Operator."""

from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import boto3
from sagemaker.estimator import Estimator
from sagemaker.inputs import TrainingInput
from sagemaker.session import Session

from zenml.client import Client
from zenml.config.build_configuration import BuildConfiguration
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.integrations.aws.flavors.sagemaker_step_operator_flavor import (
    SagemakerStepOperatorConfig,
    SagemakerStepOperatorSettings,
)
from zenml.integrations.aws.step_operators.sagemaker_step_operator_entrypoint_config import (
    SAGEMAKER_ESTIMATOR_STEP_ENV_VAR_SIZE_LIMIT,
    SagemakerStepOperatorEntrypointConfiguration,
)
from zenml.integrations.aws.utils import (
    convert_training_job_status,
)
from zenml.logger import get_logger
from zenml.orchestrators.publish_utils import publish_step_run_metadata
from zenml.stack import Stack, StackValidator
from zenml.step_operators import BaseStepOperator
from zenml.step_operators.step_operator_entrypoint_configuration import (
    StepOperatorEntrypointConfiguration,
)
from zenml.utils.env_utils import split_environment_variables
from zenml.utils.string_utils import random_str

if TYPE_CHECKING:
    from zenml.config.base_settings import BaseSettings
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import PipelineSnapshotBase, StepRunResponse

logger = get_logger(__name__)

SAGEMAKER_DOCKER_IMAGE_KEY = "sagemaker_step_operator"
_ENTRYPOINT_ENV_VARIABLE = "__ZENML_ENTRYPOINT"
STEP_JOB_NAME_METADATA_KEY = "job_name"


class SagemakerStepOperator(BaseStepOperator):
    """Step operator to run a step on Sagemaker.

    This class defines code that builds an image with the ZenML entrypoint
    to run using Sagemaker's Estimator.
    """

    @property
    def config(self) -> SagemakerStepOperatorConfig:
        """Returns the `SagemakerStepOperatorConfig` config.

        Returns:
            The configuration.
        """
        return cast(SagemakerStepOperatorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the SageMaker step operator.

        Returns:
            The settings class.
        """
        return SagemakerStepOperatorSettings

    @property
    def entrypoint_config_class(
        self,
    ) -> Type[StepOperatorEntrypointConfiguration]:
        """Returns the entrypoint configuration class for this step operator.

        Returns:
            The entrypoint configuration class for this step operator.
        """
        return SagemakerStepOperatorEntrypointConfiguration

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates the stack.

        Returns:
            A validator that checks that the stack contains a remote container
            registry and a remote artifact store.
        """

        def _validate_remote_components(stack: "Stack") -> Tuple[bool, str]:
            if stack.artifact_store.config.is_local:
                return False, (
                    "The SageMaker step operator runs code remotely and "
                    "needs to write files into the artifact store, but the "
                    f"artifact store `{stack.artifact_store.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote artifact store when using the SageMaker "
                    "step operator."
                )

            container_registry = stack.container_registry
            assert container_registry is not None

            if container_registry.config.is_local:
                return False, (
                    "The SageMaker step operator runs code remotely and "
                    "needs to push/pull Docker images, but the "
                    f"container registry `{container_registry.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote container registry when using the "
                    "SageMaker step operator."
                )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_remote_components,
        )

    def get_docker_builds(
        self, snapshot: "PipelineSnapshotBase"
    ) -> List["BuildConfiguration"]:
        """Gets the Docker builds required for the component.

        Args:
            snapshot: The pipeline snapshot for which to get the builds.

        Returns:
            The required Docker builds.
        """
        builds = []
        for step_name, step in snapshot.step_configurations.items():
            if step.config.uses_step_operator(self.name):
                build = BuildConfiguration(
                    key=SAGEMAKER_DOCKER_IMAGE_KEY,
                    settings=step.config.docker_settings,
                    step_name=step_name,
                    entrypoint=f"${_ENTRYPOINT_ENV_VARIABLE}",
                )
                builds.append(build)

        return builds

    def _get_sagemaker_session(self) -> Session:
        """Creates an authenticated Sagemaker session.

        Raises:
            RuntimeError: If the connector returns the wrong type for the
                session.

        Returns:
            The Sagemaker session.
        """
        boto_session: boto3.Session
        if connector := self.get_connector():
            boto_session = connector.connect()
            if not isinstance(boto_session, boto3.Session):
                raise RuntimeError(
                    f"Expected to receive a `boto3.Session` object from the "
                    f"linked connector, but got type `{type(boto_session)}`."
                )
        else:
            boto_session = boto3.Session()

        return Session(
            boto_session=boto_session, default_bucket=self.config.bucket
        )

    @property
    def sagemaker_session(self) -> Session:
        """Returns the Sagemaker session.

        Returns:
            The Sagemaker session.
        """
        if self.connector_has_expired():
            self._sagemaker_session = None

        if self._sagemaker_session is None:
            self._sagemaker_session = self._get_sagemaker_session()
        return self._sagemaker_session

    def submit(
        self,
        info: "StepRunInfo",
        entrypoint_command: List[str],
        environment: Dict[str, str],
    ) -> None:
        """Submits a step run to SageMaker.

        Args:
            info: Information about the step run.
            entrypoint_command: Command that executes the step.
            environment: Environment variables to set in the step operator
                environment.
        """
        if not info.config.resource_settings.empty:
            logger.warning(
                "Specifying custom step resources is not supported for "
                "the SageMaker step operator. If you want to run this step "
                "operator on specific resources, you can do so by configuring "
                "a different instance type like this: "
                "`zenml step-operator update %s "
                "--instance_type=<INSTANCE_TYPE>`",
                self.name,
            )

        settings = cast(SagemakerStepOperatorSettings, self.get_settings(info))

        if settings.environment:
            environment.update(settings.environment)

        # Sagemaker does not allow environment variables longer than 512
        # characters to be passed to Estimator steps. If an environment variable
        # is longer than 512 characters, we split it into multiple environment
        # variables (chunks) and re-construct it on the other side using the
        # custom entrypoint configuration.
        split_environment_variables(
            env=environment,
            size_limit=SAGEMAKER_ESTIMATOR_STEP_ENV_VAR_SIZE_LIMIT,
        )

        image_name = info.get_image(key=SAGEMAKER_DOCKER_IMAGE_KEY)
        environment[_ENTRYPOINT_ENV_VARIABLE] = " ".join(entrypoint_command)

        # Get and default fill SageMaker estimator arguments for full ZenML support
        estimator_args = settings.estimator_args
        estimator_args.setdefault(
            "instance_type", settings.instance_type or "ml.m5.large"
        )

        # Convert environment to a dict of strings
        environment = {key: str(value) for key, value in environment.items()}

        estimator_args["environment"] = environment
        estimator_args["instance_count"] = 1
        estimator_args["sagemaker_session"] = self.sagemaker_session

        # Create Estimator
        estimator = Estimator(image_name, self.config.role, **estimator_args)

        # SageMaker allows 63 characters at maximum for job name - ZenML uses 60 for safety margin.
        step_name = Client().get_run_step(info.step_run_id).name
        training_job_name = f"{info.pipeline.name}-{step_name}"[:55]
        suffix = random_str(4)
        unique_training_job_name = f"{training_job_name}-{suffix}"

        # Sagemaker doesn't allow any underscores in job/experiment/trial names
        sanitized_training_job_name = unique_training_job_name.replace(
            "_", "-"
        )

        # Construct training input object, if necessary
        inputs: Optional[Union[TrainingInput, Dict[str, TrainingInput]]] = None

        if isinstance(settings.input_data_s3_uri, str):
            inputs = TrainingInput(s3_data=settings.input_data_s3_uri)
        elif isinstance(settings.input_data_s3_uri, dict):
            inputs = {}
            for channel, s3_uri in settings.input_data_s3_uri.items():
                inputs[channel] = TrainingInput(s3_data=s3_uri)

        experiment_config = {}
        if settings.experiment_name:
            experiment_config = {
                "ExperimentName": settings.experiment_name,
                "TrialName": sanitized_training_job_name,
            }
        info.force_write_logs()
        estimator.fit(
            wait=False,
            inputs=inputs,
            experiment_config=experiment_config,
            job_name=sanitized_training_job_name,
        )
        publish_step_run_metadata(
            info.step_run_id,
            {
                self.id: {
                    STEP_JOB_NAME_METADATA_KEY: sanitized_training_job_name,
                }
            },
        )
        info.step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY] = (
            sanitized_training_job_name
        )

    def get_status(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Gets the status of a submitted step.

        Args:
            step_run: The step run.

        Returns:
            The step status.
        """
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        sagemaker_client = self.sagemaker_session.sagemaker_client

        status = sagemaker_client.describe_training_job(
            TrainingJobName=job_name
        )["TrainingJobStatus"]
        return convert_training_job_status(status)

    def cancel(self, step_run: "StepRunResponse") -> None:
        """Cancels a submitted step.

        Args:
            step_run: The step run.
        """
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        sagemaker_client = self.sagemaker_session.sagemaker_client
        sagemaker_client.stop_training_job(TrainingJobName=job_name)



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for code repositories."""

from typing import Optional
from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.analytics.enums import AnalyticsEvent
from zenml.analytics.utils import track_handler
from zenml.constants import (
    API,
    DEVICE_VERIFY,
    DEVICES,
    VERSION_1,
)
from zenml.enums import OAuthDeviceStatus, OnboardingStep
from zenml.models import (
    OAuthDeviceFilter,
    OAuthDeviceInternalUpdate,
    OAuthDeviceResponse,
    OAuthDeviceUpdate,
    OAuthDeviceVerificationRequest,
    Page,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    server_config,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + DEVICES,
    tags=["authorized_devices"],
    responses={401: error_response},
)


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def list_authorized_devices(
    filter_model: OAuthDeviceFilter = Depends(
        make_dependable(OAuthDeviceFilter)
    ),
    hydrate: bool = False,
    auth_context: AuthContext = Security(authorize),
) -> Page[OAuthDeviceResponse]:
    """Gets a page of OAuth2 authorized devices belonging to the current user.

    Args:
        filter_model: Filter model used for pagination, sorting,
            filtering.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.
        auth_context: The current auth context.

    Returns:
        Page of OAuth2 authorized device objects.
    """
    filter_model.set_scope_user(auth_context.user.id)
    return zen_store().list_authorized_devices(
        filter_model=filter_model, hydrate=hydrate
    )


@router.get(
    "/{device_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_authorization_device(
    device_id: UUID,
    user_code: Optional[str] = None,
    hydrate: bool = True,
    auth_context: AuthContext = Security(authorize),
) -> OAuthDeviceResponse:
    """Gets a specific OAuth2 authorized device using its unique ID.

    Args:
        device_id: The ID of the OAuth2 authorized device to get.
        user_code: The user code of the OAuth2 authorized device to get. Needs
            to be specified with devices that have not been verified yet.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.
        auth_context: The current auth context.

    Returns:
        A specific OAuth2 authorized device object.

    Raises:
        KeyError: If the device with the given ID does not exist, does not
            belong to the current user or could not be verified using the
            given user code.
    """
    device = zen_store().get_authorized_device(
        device_id=device_id, hydrate=hydrate
    )
    if not device.user_id:
        # A device that hasn't been verified and associated with a user yet
        # can only be retrieved if the user code is specified and valid.
        if user_code:
            internal_device = zen_store().get_internal_authorized_device(
                device_id=device_id, hydrate=hydrate
            )
            if internal_device.verify_user_code(user_code=user_code):
                return device
    elif device.user_id == auth_context.user.id:
        return device

    raise KeyError(
        f"Unable to get device with ID {device_id}: No device with "
        "this ID found."
    )


@router.put(
    "/{device_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_authorized_device(
    device_id: UUID,
    update: OAuthDeviceUpdate,
    auth_context: AuthContext = Security(authorize),
) -> OAuthDeviceResponse:
    """Updates a specific OAuth2 authorized device using its unique ID.

    Args:
        device_id: The ID of the OAuth2 authorized device to update.
        update: The model containing the attributes to update.
        auth_context: The current auth context.

    Returns:
        The updated OAuth2 authorized device object.

    Raises:
        KeyError: If the device with the given ID does not exist or does not
            belong to the current user.
    """
    device = zen_store().get_authorized_device(device_id=device_id)
    if not device.user_id or device.user_id != auth_context.user.id:
        raise KeyError(
            f"Unable to get device with ID {device_id}: No device with "
            "this ID found."
        )

    return zen_store().update_authorized_device(
        device_id=device_id, update=update
    )


@router.put(
    "/{device_id}" + DEVICE_VERIFY,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def verify_authorized_device(
    device_id: UUID,
    request: OAuthDeviceVerificationRequest,
    auth_context: AuthContext = Security(authorize),
) -> OAuthDeviceResponse:
    """Verifies a specific OAuth2 authorized device using its unique ID.

    This endpoint implements the OAuth2 device authorization grant flow as
    defined in https://tools.ietf.org/html/rfc8628. It is called to verify
    the user code for a given device ID.

    If the user code is valid, the device is marked as verified and associated
    with the user that authorized the device. This association is required to
    be able to issue access tokens or revoke the device later on.

    Args:
        device_id: The ID of the OAuth2 authorized device to update.
        request: The model containing the verification request.
        auth_context: The current auth context.

    Returns:
        The updated OAuth2 authorized device object.

    Raises:
        ValueError: If the device verification request fails.
    """
    config = server_config()
    store = zen_store()

    # Check if a device is registered for the ID
    device_model = store.get_internal_authorized_device(
        device_id=device_id,
    )

    with track_handler(event=AnalyticsEvent.DEVICE_VERIFIED):
        # Check if the device is in a state that allows verification.
        if device_model.status != OAuthDeviceStatus.PENDING:
            raise ValueError(
                "Invalid request: device not pending verification.",
            )

        # Check if the device verification has expired.
        if device_model.expires and device_model.expires < utc_now():
            raise ValueError(
                "Invalid request: device verification expired.",
            )

        # Check if the device already has a user associated with it. If so, the
        # current user and the user associated with the device must be the same.
        if (
            device_model.user_id
            and device_model.user_id != auth_context.user.id
        ):
            raise ValueError(
                "Invalid request: this device is associated with another user.",
            )

        # Check if the device code is valid.
        if not device_model.verify_user_code(request.user_code):
            # If the device code is invalid, increment the failed auth attempts
            # counter and lock the device if the maximum number of failed auth
            # attempts has been reached.
            failed_auth_attempts = device_model.failed_auth_attempts + 1
            update = OAuthDeviceInternalUpdate(
                failed_auth_attempts=failed_auth_attempts
            )
            if failed_auth_attempts >= config.max_failed_device_auth_attempts:
                update.locked = True
            store.update_internal_authorized_device(
                device_id=device_model.id,
                update=update,
            )
            if failed_auth_attempts >= config.max_failed_device_auth_attempts:
                raise ValueError(
                    "Invalid request: device locked due to too many failed "
                    "authentication attempts.",
                )
            raise ValueError(
                "Invalid request: invalid user code.",
            )

        # If the device code is valid, associate the device with the current user.
        # We don't reset the expiration date yet, because we want to make sure
        # that the client has received the access token before we do so, to avoid
        # brute force attacks on the device code.
        update = OAuthDeviceInternalUpdate(
            status=OAuthDeviceStatus.VERIFIED,
            user_id=auth_context.user.id,
            failed_auth_attempts=0,
            trusted_device=request.trusted_device,
        )
        device_model = store.update_internal_authorized_device(
            device_id=device_model.id,
            update=update,
        )

    store.update_onboarding_state(
        completed_steps={OnboardingStep.DEVICE_VERIFIED}
    )
    return device_model


@router.delete(
    "/{device_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_authorized_device(
    device_id: UUID,
    auth_context: AuthContext = Security(authorize),
) -> None:
    """Deletes a specific OAuth2 authorized device using its unique ID.

    Args:
        device_id: The ID of the OAuth2 authorized device to delete.
        auth_context: The current auth context.

    Raises:
        KeyError: If the device with the given ID does not exist or does not
            belong to the current user.
    """
    device = zen_store().get_authorized_device(device_id=device_id)
    if not device.user_id or device.user_id != auth_context.user.id:
        raise KeyError(
            f"Unable to get device with ID {device_id}: No device with "
            "this ID found."
        )

    zen_store().delete_authorized_device(device_id=device_id)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality to administer users of the ZenML CLI and server."""

from typing import Any, Optional

import click

from zenml.cli import utils as cli_utils
from zenml.cli.cli import TagGroup, cli
from zenml.cli.utils import (
    OutputFormat,
    is_sorted_or_filtered,
    list_options,
)
from zenml.client import Client
from zenml.config.global_config import GlobalConfiguration
from zenml.console import console
from zenml.enums import CliCategories, StoreType
from zenml.exceptions import (
    AuthorizationException,
    EntityExistsError,
    IllegalOperationError,
)
from zenml.models import UserFilter


@cli.group(cls=TagGroup, tag=CliCategories.IDENTITY_AND_SECURITY)
def user() -> None:
    """Commands for user management."""


@user.command("describe")
@click.argument("user_name_or_id", type=str, required=False)
def describe_user(user_name_or_id: Optional[str] = None) -> None:
    """Get the user.

    Args:
        user_name_or_id: The name or ID of the user.
    """
    client = Client()
    if not user_name_or_id:
        active_user = client.active_user
        cli_utils.print_pydantic_models(
            [active_user],
            exclude_columns=[
                "created",
                "updated",
                "email",
                "email_opted_in",
                "activation_token",
            ],
        )
    else:
        try:
            user = client.get_user(user_name_or_id)
        except KeyError as err:
            cli_utils.exception(err)
        else:
            cli_utils.print_pydantic_models(
                [user],
                exclude_columns=[
                    "created",
                    "updated",
                    "email",
                    "email_opted_in",
                    "activation_token",
                ],
            )


@user.command("list")
@list_options(
    UserFilter, default_columns=["active", "id", "name", "full_name"]
)
@click.pass_context
def list_users(
    ctx: click.Context,
    /,
    columns: str,
    output_format: OutputFormat,
    **kwargs: Any,
) -> None:
    """List all users.

    Args:
        ctx: The click context object
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        kwargs: Keyword arguments to filter the list of users.
    """
    client = Client()
    with console.status("Listing users...\n"):
        users = client.list_users(**kwargs)

    # Handle active user highlighting (only if not filtered/sorted)
    show_active = not is_sorted_or_filtered(ctx)
    if show_active and users.items:
        active_user_id = client.active_user.id
        if active_user_id not in {u.id for u in users.items}:
            users.items.insert(0, client.active_user)
        users.items.sort(key=lambda u: u.id != active_user_id)
    else:
        active_user_id = None

    cli_utils.print_page(
        users,
        columns,
        output_format,
        empty_message="No users found for the given filters.",
        row_generator=cli_utils.generate_user_row,
        active_id=active_user_id,
    )


@user.command(
    "create",
    help="Create a new user. If an empty password is configured, an activation "
    "token is generated and a link to the dashboard is provided where the "
    "user can activate the account.",
)
@click.argument("user_name", type=str, required=True)
@click.option(
    "--password",
    help=(
        "The user password. If omitted, a prompt will be shown to enter the "
        "password. If an empty password is entered, an activation token is "
        "generated and a link to the dashboard is provided where the user can "
        "activate the account."
    ),
    required=False,
    type=str,
)
@click.option(
    "--is_admin",
    is_flag=True,
    help=(
        "Whether the user should be an admin. If not specified, the user will "
        "be a regular user."
    ),
    required=False,
    default=False,
)
def create_user(
    user_name: str,
    password: Optional[str] = None,
    is_admin: bool = False,
) -> None:
    """Create a new user.

    Args:
        user_name: The name of the user to create.
        password: The password of the user to create.
        is_admin: Whether the user should be an admin.
    """
    client = Client()
    if not password:
        if client.zen_store.type != StoreType.REST:
            password = click.prompt(
                f"Password for user {user_name}",
                hide_input=True,
            )
        else:
            password = click.prompt(
                f"Password for user {user_name}. Leave empty to generate an "
                f"activation token",
                default="",
                hide_input=True,
            )
    else:
        cli_utils.warning(
            "Supplying password values in the command line is not safe. "
            "Please consider using the prompt option."
        )

    try:
        new_user = client.create_user(
            name=user_name, password=password, is_admin=is_admin
        )

        cli_utils.declare(f"Created user '{new_user.name}'.")
    except EntityExistsError as err:
        cli_utils.exception(err)
    else:
        if not new_user.active and new_user.activation_token is not None:
            user_info = f"?user={str(new_user.id)}&username={new_user.name}&token={new_user.activation_token}"
            cli_utils.declare(
                f"The created user account is currently inactive. You can "
                f"activate it by visiting the dashboard at the following URL:\n"
                # TODO: keep only `activate-user` once legacy dashboard is gone
                f"{client.zen_store.url}/activate-user{user_info}\n\n"
                "If you are using Legacy dashboard visit the following URL:\n"
                f"{client.zen_store.url}/signup{user_info}\n"
            )


@user.command(
    "update",
    help="Update user information through the cli.",
)
@click.argument("user_name_or_id", type=str, required=True)
@click.option(
    "--name",
    "-n",
    "updated_name",
    type=str,
    required=False,
    help="New user name.",
)
@click.option(
    "--full_name",
    "-f",
    "updated_full_name",
    type=str,
    required=False,
    help="New full name. If this contains an empty space make sure to surround "
    "the name with quotes '<Full Name>'.",
)
@click.option(
    "--email",
    "-e",
    "updated_email",
    type=str,
    required=False,
    help="New user email.",
)
@click.option(
    "--admin",
    "-a",
    "make_admin",
    is_flag=True,
    required=False,
    default=None,
    help="Whether the user should be an admin.",
)
@click.option(
    "--user",
    "-u",
    "make_user",
    is_flag=True,
    required=False,
    default=None,
    help="Whether the user should be a regular user.",
)
@click.option(
    "--active",
    "active",
    type=bool,
    required=False,
    default=None,
    help="Use to activate or deactivate a user account.",
)
def update_user(
    user_name_or_id: str,
    updated_name: Optional[str] = None,
    updated_full_name: Optional[str] = None,
    updated_email: Optional[str] = None,
    make_admin: Optional[bool] = None,
    make_user: Optional[bool] = None,
    active: Optional[bool] = None,
) -> None:
    """Update an existing user.

    Args:
        user_name_or_id: The name of the user to create.
        updated_name: The name of the user to create.
        updated_full_name: The name of the user to create.
        updated_email: The name of the user to create.
        make_admin: Whether the user should be an admin.
        make_user: Whether the user should be a regular user.
        active: Use to activate or deactivate a user account.
    """
    if make_admin is not None and make_user is not None:
        cli_utils.error(
            "Cannot set both --admin and --user flags as these are mutually exclusive."
        )
    try:
        current_user = Client().get_user(
            user_name_or_id, allow_name_prefix_match=False
        )
        if current_user.is_admin and make_user:
            confirmation = cli_utils.confirmation(
                f"Currently user `{current_user.name}` is an admin. Are you "
                "sure you want to make them a regular user?"
            )
            if not confirmation:
                cli_utils.declare("User update canceled.")
                return

        updated_is_admin = None
        if make_admin is True:
            updated_is_admin = True
        elif make_user is True:
            updated_is_admin = False
        Client().update_user(
            name_id_or_prefix=user_name_or_id,
            updated_name=updated_name,
            updated_full_name=updated_full_name,
            updated_email=updated_email,
            updated_is_admin=updated_is_admin,
            active=active,
        )
    except (KeyError, IllegalOperationError) as err:
        cli_utils.exception(err)


@user.command(
    "change-password",
    help="Change the password for the current user account.",
)
@click.option(
    "--password",
    help=(
        "The new user password. If omitted, a prompt will be shown to enter "
        "the password."
    ),
    required=False,
    type=str,
)
@click.option(
    "--old-password",
    help=(
        "The old user password. If omitted, a prompt will be shown to enter "
        "the old password."
    ),
    required=False,
    type=str,
)
def change_user_password(
    password: Optional[str] = None, old_password: Optional[str] = None
) -> None:
    """Change the password of the current user.

    Args:
        password: The new password for the current user.
        old_password: The old password for the current user.
    """
    active_user = Client().active_user

    if old_password is not None or password is not None:
        cli_utils.warning(
            "Supplying password values in the command line is not safe. "
            "Please consider using the prompt option."
        )

    if old_password is None:
        old_password = click.prompt(
            f"Current password for user {active_user.name}",
            hide_input=True,
        )
    if password is None:
        password = click.prompt(
            f"New password for user {active_user.name}",
            hide_input=True,
        )
        password_again = click.prompt(
            f"Please re-enter the new password for user {active_user.name}",
            hide_input=True,
        )
        if password != password_again:
            cli_utils.error("Passwords do not match.")

    try:
        Client().update_user(
            name_id_or_prefix=active_user.id,
            old_password=old_password,
            updated_password=password,
        )
    except (KeyError, IllegalOperationError, AuthorizationException) as err:
        cli_utils.exception(err)

    cli_utils.declare(
        f"Successfully updated password for active user '{active_user.name}'."
    )


@user.command(
    "deactivate",
    help="Generate an activation token to reset the password for a user account",
)
@click.argument("user_name_or_id", type=str, required=True)
def deactivate_user(
    user_name_or_id: str,
) -> None:
    """Reset the password of a user.

    Args:
        user_name_or_id: The name or ID of the user to reset the password for.
    """
    client = Client()

    store = GlobalConfiguration().store_configuration
    if store.type != StoreType.REST:
        cli_utils.error(
            "Deactivating users is only supported when connected to a ZenML "
            "server."
        )

    try:
        if not client.active_user.is_admin:
            cli_utils.error(
                "Only admins can reset the password of other users."
            )

        user = client.deactivate_user(
            name_id_or_prefix=user_name_or_id,
        )
    except (KeyError, IllegalOperationError) as err:
        cli_utils.exception(err)

    user_info = f"?user={str(user.id)}&username={user.name}&token={user.activation_token}"
    cli_utils.declare(
        f"Successfully deactivated user account '{user.name}'."
        f"To reactivate the account, please visit the dashboard at the "
        "following URL:\n"
        # TODO: keep only `activate-user` once legacy dashboard is gone
        f"{client.zen_store.url}/activate-user{user_info}\n\n"
        "If you are using Legacy dashboard visit the following URL:\n"
        f"{client.zen_store.url}/signup{user_info}\n"
    )


@user.command("delete")
@click.argument("user_name_or_id", type=str, required=True)
def delete_user(user_name_or_id: str) -> None:
    """Delete a user.

    Args:
        user_name_or_id: The name or ID of the user to delete.
    """
    try:
        Client().delete_user(user_name_or_id)
    except (KeyError, IllegalOperationError) as err:
        cli_utils.exception(err)
    cli_utils.declare(f"Deleted user '{user_name_or_id}'.")



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utilities for creating step runs."""

import json
from datetime import timedelta
from typing import Dict, List, Optional, Set, Tuple
from uuid import UUID, uuid4

from zenml.client import Client
from zenml.config.step_configurations import Step
from zenml.constants import (
    CODE_HASH_PARAMETER_NAME,
    ENV_ZENML_STEP_RUNS_FETCH_MAX_CHUNK_LENGTH,
    TEXT_FIELD_MAX_LENGTH,
    handle_int_env_var,
)
from zenml.enums import ExecutionStatus
from zenml.logger import get_logger
from zenml.model.utils import link_artifact_version_to_model_version
from zenml.models import (
    ArtifactVersionResponse,
    ModelVersionResponse,
    PipelineRunResponse,
    PipelineSnapshotResponse,
    StepRunRequest,
    StepRunResponse,
)
from zenml.orchestrators import cache_utils, input_utils, utils
from zenml.stack import Stack
from zenml.utils import pagination_utils
from zenml.utils.time_utils import utc_now

logger = get_logger(__name__)


class StepRunRequestFactory:
    """Helper class to create step run requests."""

    def __init__(
        self,
        snapshot: "PipelineSnapshotResponse",
        pipeline_run: "PipelineRunResponse",
        stack: "Stack",
    ) -> None:
        """Initialize the object.

        Args:
            snapshot: The snapshot for which to create step run requests.
            pipeline_run: The pipeline run for which to create step run
                requests.
            stack: The stack on which the pipeline run is happening.
        """
        self.snapshot = snapshot
        self.pipeline_run = pipeline_run
        self.stack = stack
        self._original_step_run_cache: Dict[
            str, Optional["StepRunResponse"]
        ] = {}

    def _get_original_step_run(
        self, invocation_id: str
    ) -> Optional["StepRunResponse"]:
        """Get the original step run for a replayed invocation.

        Args:
            invocation_id: The invocation ID to look up.

        Returns:
            The corresponding step run from the original pipeline run, if it
            exists.
        """
        if invocation_id in self._original_step_run_cache:
            return self._original_step_run_cache[invocation_id]

        if not self.pipeline_run.original_run:
            return None

        step_runs = Client().list_run_steps(
            pipeline_run_id=self.pipeline_run.original_run.id,
            name=invocation_id,
            size=1,
            hydrate=True,
            exclude_retried=True,
        )
        original_step_run = step_runs[0] if step_runs else None
        self._original_step_run_cache[invocation_id] = original_step_run
        return original_step_run

    def should_skip_step(self, invocation_id: str) -> bool:
        """Check whether a step should be skipped.

        Args:
            invocation_id: The invocation ID to check.

        Returns:
            Whether the step should be skipped.
        """
        if invocation_id in self.snapshot.pipeline_configuration.steps_to_skip:
            return True

        if not self.snapshot.pipeline_configuration.skip_successful_steps:
            return False

        if self._get_input_overrides(invocation_id):
            logger.warning(
                "Step `%s` should be skipped, but there are input overrides "
                "configured. The step will be executed.",
                invocation_id,
            )
            return False

        try:
            original_step_run = self._get_original_step_run(invocation_id)
        except Exception as e:
            logger.warning(
                "Failed to fetch original step run `%s` while trying to skip "
                "successful steps: %s. The step will be executed.",
                invocation_id,
                str(e),
            )
            return False

        return bool(
            original_step_run and original_step_run.status.is_successful
        )

    def has_caching_enabled(self, invocation_id: str) -> bool:
        """Check if the step has caching enabled.

        Args:
            invocation_id: The invocation ID for which to check if caching is
                enabled.

        Returns:
            Whether the step has caching enabled.
        """
        step = self.snapshot.step_configurations[invocation_id]
        return utils.is_setting_enabled(
            is_enabled_on_step=step.config.enable_cache,
            is_enabled_on_pipeline=self.snapshot.pipeline_configuration.enable_cache,
        )

    def create_request(
        self, invocation_id: str, dynamic_config: Optional[Step] = None
    ) -> StepRunRequest:
        """Create a step run request.

        This will only create a request with basic information and will not yet
        compute information like the cache key and inputs. This is separated
        into a different method `populate_request(...)` that might raise
        exceptions while trying to compute this information.

        Args:
            invocation_id: The invocation ID for which to create the request.
            dynamic_config: The dynamic configuration for the step.

        Returns:
            The step run request.
        """
        return StepRunRequest(
            name=invocation_id,
            pipeline_run_id=self.pipeline_run.id,
            status=ExecutionStatus.RUNNING,
            start_time=utc_now(),
            project=Client().active_project.id,
            dynamic_config=dynamic_config,
        )

    def populate_request(
        self,
        request: StepRunRequest,
        step_runs: Optional[Dict[str, "StepRunResponse"]] = None,
    ) -> None:
        """Populate a step run request with additional information.

        Args:
            request: The request to populate.
            step_runs: A dictionary of already fetched step runs to use for
                input resolution. This will be updated in-place with newly
                fetched step runs.
        """
        if self.should_skip_step(request.name):
            self._populate_skipped_step(request)
            return

        step = (
            request.dynamic_config
            or self.snapshot.step_configurations[request.name]
        )
        input_overrides = self._get_input_overrides(request.name, step=step)

        input_artifacts = input_utils.resolve_step_inputs(
            step=step,
            pipeline_run=self.pipeline_run,
            step_runs=step_runs,
            input_overrides=input_overrides,
        )

        request.inputs = {
            name: [artifact.id for artifact in artifacts]
            for name, artifacts in input_artifacts.items()
        }

        cache_key = cache_utils.generate_cache_key(
            step=step,
            input_artifacts=input_artifacts,
            artifact_store=self.stack.artifact_store,
            project_id=Client().active_project.id,
        )
        request.cache_key = cache_key

        if step.config.cache_policy.expires_after:
            request.cache_expires_at = request.start_time + timedelta(
                seconds=step.config.cache_policy.expires_after
            )

        (
            docstring,
            source_code,
        ) = self._get_docstring_and_source_code(step=step)

        request.docstring = docstring
        request.source_code = source_code
        request.code_hash = step.config.parameters.get(
            CODE_HASH_PARAMETER_NAME
        )

        cache_enabled = utils.is_setting_enabled(
            is_enabled_on_step=step.config.enable_cache,
            is_enabled_on_pipeline=self.snapshot.pipeline_configuration.enable_cache,
        )

        if cache_enabled:
            if cached_step_run := cache_utils.get_cached_step_run(
                cache_key=cache_key
            ):
                request.inputs = {
                    input_name: [artifact.id for artifact in artifacts]
                    for input_name, artifacts in cached_step_run.inputs.items()
                }

                request.original_step_run_id = cached_step_run.id
                request.outputs = {
                    output_name: [artifact.id for artifact in artifacts]
                    for output_name, artifacts in cached_step_run.outputs.items()
                }

                request.status = ExecutionStatus.CACHED
                request.end_time = request.start_time

                # Cached step runs themselves are not valid candidates for
                # caching, but just in case we set the cache expiration time to
                # the time when the cached step run expires.
                request.cache_expires_at = cached_step_run.cache_expires_at

                # As a last resort, we try to reuse the docstring/source code
                # from the cached step run. This is part of the cache key
                # computation, so it must be identical to the one we would have
                # computed ourselves.
                if request.source_code is None:
                    request.source_code = cached_step_run.source_code
                if request.docstring is None:
                    request.docstring = cached_step_run.docstring

    def _populate_skipped_step(
        self,
        request: StepRunRequest,
    ) -> None:
        """Populate a skipped step run request.

        Args:
            request: The request to populate.

        Raises:
            RuntimeError: If the pipeline run is not a replayed run.
            RuntimeError: If no step run is found for the step in the original
                run.
            RuntimeError: If the step wasn't successfully completed in the
                original run.
        """
        if not self.pipeline_run.original_run:
            raise RuntimeError(
                "Unable to populate skipped step run request because the "
                "pipeline run is not a replayed run."
            )

        original_step_run = self._get_original_step_run(request.name)
        if not original_step_run:
            raise RuntimeError(
                f"No step run found for step `{request.name}` in original run "
                f"`{self.pipeline_run.original_run.id}`."
            )
        if not original_step_run.status.is_successful:
            raise RuntimeError(
                f"Step `{request.name}` cannot be skipped because it wasn't "
                "successfully completed in the original run."
            )

        request.status = ExecutionStatus.SKIPPED
        request.original_step_run_id = original_step_run.id
        request.end_time = request.start_time

        request.inputs = {
            input_name: [artifact.id for artifact in artifacts]
            for input_name, artifacts in original_step_run.inputs.items()
        }
        request.outputs = {
            output_name: [artifact.id for artifact in artifacts]
            for output_name, artifacts in original_step_run.outputs.items()
        }

        # Skipped step runs themselves are not valid candidates for
        # caching, but just in case we set the cache to expire immediately.
        request.cache_key = str(uuid4())
        request.cache_expires_at = request.start_time

        request.code_hash = original_step_run.code_hash
        request.source_code = original_step_run.source_code
        request.docstring = original_step_run.docstring

    def _get_docstring_and_source_code(
        self, step: "Step"
    ) -> Tuple[Optional[str], Optional[str]]:
        """Get the docstring and source code for the step.

        Args:
            step: The step for which to get the docstring and source code.

        Returns:
            The docstring and source code of the step.
        """
        try:
            return self._get_docstring_and_source_code_from_step_instance(
                step=step
            )
        except ImportError:
            pass

        # Failed to import the step instance, this is most likely because this
        # code is running on the server as part of a template execution.
        # We now try to fetch the docstring/source code from a step run of the
        # snapshot that was used to create the template
        return self._try_to_get_docstring_and_source_code_from_template(
            invocation_id=step.spec.invocation_id
        )

    @staticmethod
    def _get_docstring_and_source_code_from_step_instance(
        step: "Step",
    ) -> Tuple[Optional[str], str]:
        """Get the docstring and source code of a step.

        Args:
            step: The step configuration.

        Returns:
            The docstring and source code of a step.
        """
        from zenml.steps.base_step import BaseStep

        step_instance = BaseStep.load_from_source(step.spec.source)

        docstring = step_instance.docstring
        if docstring and len(docstring) > TEXT_FIELD_MAX_LENGTH:
            docstring = docstring[: (TEXT_FIELD_MAX_LENGTH - 3)] + "..."

        source_code = step_instance.source_code
        if source_code and len(source_code) > TEXT_FIELD_MAX_LENGTH:
            source_code = source_code[: (TEXT_FIELD_MAX_LENGTH - 3)] + "..."

        return docstring, source_code

    def _try_to_get_docstring_and_source_code_from_template(
        self, invocation_id: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Try to get the docstring and source code via a potential template.

        Args:
            invocation_id: The step invocation ID for which to get the
                docstring and source code.

        Returns:
            The docstring and source code of the step.
        """
        if template_id := self.pipeline_run.template_id:
            template = Client().get_run_template(template_id)
            if (
                snapshot_id := template.source_snapshot.id
                if template.source_snapshot
                else None
            ):
                steps = Client().list_run_steps(
                    snapshot_id=snapshot_id,
                    name=invocation_id,
                    size=1,
                    hydrate=True,
                )

                if len(steps) > 0:
                    step = steps[0]
                    return step.docstring, step.source_code

        return None, None

    def _get_input_overrides(
        self, invocation_id: str, step: Optional["Step"] = None
    ) -> Dict[str, "UUID"]:
        """Get input overrides for a step.

        Args:
            invocation_id: The invocation ID to look up.
            step: The step configuration. If not provided, no input key
                validation is performed.

        Returns:
            The input overrides for the step.
        """
        overrides = (
            self.snapshot.pipeline_configuration.step_input_overrides.get(
                invocation_id, {}
            )
        )

        if step:
            available_input_keys = step.available_input_keys
            invalid_keys = overrides.keys() - available_input_keys
            if invalid_keys:
                logger.warning(
                    "Ignoring invalid input overrides for step `%s`: %s",
                    invocation_id,
                    invalid_keys,
                )

            overrides = {
                key: value
                for key, value in overrides.items()
                if key in available_input_keys
            }

        return overrides


def find_cacheable_invocation_candidates(
    snapshot: "PipelineSnapshotResponse",
    finished_invocations: Set[str],
) -> Set[str]:
    """Find invocations that can potentially be cached.

    Args:
        snapshot: The pipeline snapshot containing the invocations.
        finished_invocations: A set of invocations that are already finished.

    Returns:
        The set of invocations that can potentially be cached.
    """
    invocations = set()
    for invocation_id, step in snapshot.step_configurations.items():
        if invocation_id in finished_invocations:
            continue

        cache_enabled = utils.is_setting_enabled(
            is_enabled_on_step=step.config.enable_cache,
            is_enabled_on_pipeline=snapshot.pipeline_configuration.enable_cache,
        )

        if not cache_enabled:
            continue

        if set(step.spec.upstream_steps) - finished_invocations:
            continue

        invocations.add(invocation_id)

    return invocations


def create_cached_step_runs(
    snapshot: "PipelineSnapshotResponse",
    pipeline_run: PipelineRunResponse,
    stack: "Stack",
) -> Set[str]:
    """Create all cached step runs for a pipeline run.

    Args:
        snapshot: The snapshot of the pipeline run.
        pipeline_run: The pipeline run for which to create the step runs.
        stack: The stack on which the pipeline run is happening.

    Returns:
        The invocation IDs of the created step runs.
    """
    cached_invocations: Set[str] = set()
    visited_invocations: Set[str] = set()
    request_factory = StepRunRequestFactory(
        snapshot=snapshot, pipeline_run=pipeline_run, stack=stack
    )
    # This is used to cache the step runs that we created to avoid unnecessary
    # server requests.
    step_runs: Dict[str, "StepRunResponse"] = {}

    while (
        cache_candidates := find_cacheable_invocation_candidates(
            snapshot=snapshot,
            finished_invocations=cached_invocations,
        )
        # We've already checked these invocations and were not able to cache
        # them -> no need to check them again
        - visited_invocations
    ):
        for invocation_id in cache_candidates:
            visited_invocations.add(invocation_id)

            try:
                step_run_request = request_factory.create_request(
                    invocation_id
                )
                request_factory.populate_request(
                    step_run_request, step_runs=step_runs
                )
            except Exception as e:
                # We failed to create/populate the step run. This might be due
                # to some input resolution error, or an error importing the step
                # source (there might be some missing dependencies). We just
                # treat the step as not cached and let the orchestrator try
                # again in a potentially different environment.
                logger.debug(
                    "Failed preparing step run `%s`: %s", invocation_id, str(e)
                )
                continue

            if step_run_request.status != ExecutionStatus.CACHED:
                # If we're not able to cache the step run, the orchestrator
                # will run the step later which will create the step run
                # -> We don't need to do anything here
                continue

            step_run = publish_cached_step_run(
                step_run_request, pipeline_run=pipeline_run
            )

            # Include the newly created step run in the step runs dictionary to
            # avoid fetching it again later when downstream steps need it for
            # input resolution.
            step_runs[invocation_id] = step_run

            logger.info("Using cached version of step `%s`.", invocation_id)
            cached_invocations.add(invocation_id)

    return cached_invocations


def log_model_version_dashboard_url(
    model_version: ModelVersionResponse,
) -> None:
    """Log the dashboard URL for a model version.

    If the current server is not a ZenML Pro workspace, a fallback message is
    logged instead.

    Args:
        model_version: The model version for which to log the dashboard URL.
    """
    from zenml.utils.dashboard_utils import get_model_version_url

    if model_version_url := get_model_version_url(model_version):
        logger.info(
            "Dashboard URL for Model Version `%s (%s)`:\n%s",
            model_version.model.name,
            model_version.name,
            model_version_url,
        )
    else:
        logger.info(
            "Models can be viewed in the dashboard using ZenML Pro. Sign up "
            "for a free trial at https://www.zenml.io/pro/"
        )


def link_output_artifacts_to_model_version(
    artifacts: Dict[str, List[ArtifactVersionResponse]],
    model_version: ModelVersionResponse,
) -> None:
    """Link the outputs of a step run to a model version.

    Args:
        artifacts: The step output artifacts.
        model_version: The model version to link.
    """
    for output_artifacts in artifacts.values():
        for output_artifact in output_artifacts:
            link_artifact_version_to_model_version(
                artifact_version=output_artifact,
                model_version=model_version,
            )


def publish_cached_step_run(
    request: "StepRunRequest", pipeline_run: "PipelineRunResponse"
) -> "StepRunResponse":
    """Create a cached step run and link to model version and tags.

    Args:
        request: The request for the step run.
        pipeline_run: The pipeline run of the step.

    Returns:
        The createdstep run.
    """
    step_run = Client().zen_store.create_run_step(request)

    if model_version := step_run.model_version or pipeline_run.model_version:
        link_output_artifacts_to_model_version(
            artifacts=step_run.outputs,
            model_version=model_version,
        )

    return step_run


def fetch_step_runs_by_names(
    step_run_names: List[str], pipeline_run: "PipelineRunResponse"
) -> Dict[str, "StepRunResponse"]:
    """Fetch step runs by names.

    Args:
        step_run_names: The names of the step runs to fetch.
        pipeline_run: The pipeline run of the step runs.

    Returns:
        A dictionary of step runs by name.
    """
    step_runs = {}

    chunks = []
    current_chunk = []
    current_length = 0
    max_chunk_length = handle_int_env_var(
        ENV_ZENML_STEP_RUNS_FETCH_MAX_CHUNK_LENGTH, default=5000
    )

    for step_name in step_run_names:
        current_chunk.append(step_name)
        current_length += len(step_name) + 5  # 5 is for the JSON encoding

        if current_length > max_chunk_length:
            chunks.append(current_chunk)
            current_chunk = []
            current_length = 0

    if current_chunk:
        chunks.append(current_chunk)

    for chunk in chunks:
        step_runs.update(
            {
                run_step.name: run_step
                for run_step in pagination_utils.depaginate(
                    Client().list_run_steps,
                    pipeline_run_id=pipeline_run.id,
                    project=pipeline_run.project_id,
                    name="oneof:" + json.dumps(chunk),
                )
            }
        )
    return step_runs



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Context variable utilities."""

import contextvars
import threading
from contextvars import ContextVar
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    List,
    Optional,
    TypeVar,
)

from typing_extensions import ParamSpec, Self

T = TypeVar("T")

P = ParamSpec("P")
R = TypeVar("R")


class BaseContext:
    """Base context class."""

    __context_var__: ClassVar[contextvars.ContextVar[Self]]

    def __init__(self) -> None:
        """Initialize the context."""
        self._token: Optional[contextvars.Token[Any]] = None

    @classmethod
    def get(cls: type[Self]) -> Optional[Self]:
        """Get the active context.

        Returns:
            The active context.
        """
        return cls.__context_var__.get(None)

    @classmethod
    def is_active(cls: type[Self]) -> bool:
        """Check if the context is active.

        Returns:
            True if the context is active, False otherwise.
        """
        return cls.get() is not None

    def __enter__(self) -> Self:
        """Enter the context.

        Returns:
            The context object.
        """
        self._token = self.__context_var__.set(self)
        return self

    def __exit__(self, *_: Any) -> None:
        """Exit the context.

        Args:
            *_: Unused keyword arguments.

        Raises:
            RuntimeError: If the context has not been entered.
        """
        if not self._token:
            raise RuntimeError(
                f"Can't exit {self.__class__.__name__} because it has not been "
                "entered."
            )
        self.__context_var__.reset(self._token)


class ContextVarList(Generic[T]):
    """Thread-safe wrapper around ContextVar[List] with atomic add/remove operations."""

    def __init__(self, name: str) -> None:
        """Initialize the context variable list.

        Args:
            name: The name for the underlying ContextVar.
        """
        # Use None as default to avoid mutable default issues
        self._context_var: ContextVar[Optional[List[T]]] = ContextVar(
            name, default=None
        )
        # Lock to ensure atomic operations
        self._lock = threading.Lock()

    def get(self) -> List[T]:
        """Get the current list value. Returns empty list if not set.

        Returns:
            The current list value.
        """
        value = self._context_var.get()
        return value if value is not None else []

    def add(self, item: T) -> None:
        """Thread-safely add an item to the list.

        Args:
            item: The item to add to the list.
        """
        with self._lock:
            current_list = self.get()
            if not any(x is item for x in current_list):
                new_list = current_list + [item]
                self._context_var.set(new_list)

    def remove(self, item: T) -> None:
        """Thread-safely remove an item from the list.

        Args:
            item: The item to remove from the list.
        """
        with self._lock:
            current_list = self.get()
            if any(x is item for x in current_list):
                new_list = [x for x in current_list if x is not item]
                self._context_var.set(new_list)


def run_in_current_context(
    func: Callable[P, R], *args: P.args, **kwargs: P.kwargs
) -> R:
    """Run a function in the current context.

    Args:
        func: The function to run.
        *args: The arguments to pass to the function.
        **kwargs: The keyword arguments to pass to the function.

    Returns:
        The result of the function.
    """
    ctx = contextvars.copy_context()
    return ctx.run(func, *args, **kwargs)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base app extension interface."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from zenml.deployers.server.app import BaseDeploymentAppRunner


class BaseAppExtension(ABC):
    """Abstract base for app extensions.

    Extensions provide advanced framework-specific capabilities like:
    - Custom authentication/authorization
    - Observability (logging, tracing, metrics)
    - Complex routers with framework-specific features
    - OpenAPI customizations
    - Advanced middleware patterns

    Subclasses must implement install() to modify the app.
    """

    @abstractmethod
    def install(
        self,
        app_runner: "BaseDeploymentAppRunner",
    ) -> None:
        """Install extension into the application.

        Args:
            app_runner: The deployment app runner instance being used to build
                and run the web application.

        Raises:
            RuntimeError: If installation fails.
        """



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the MLflow experiment tracker for ZenML."""

import importlib
import os
from typing import TYPE_CHECKING, Any, Dict, Optional, Type, cast

import mlflow
from mlflow.entities import Experiment, LifecycleStage, Run
from mlflow.exceptions import MlflowException
from mlflow.store.db.db_types import DATABASE_ENGINES

import zenml
from zenml.artifact_stores import LocalArtifactStore
from zenml.client import Client
from zenml.config.base_settings import BaseSettings
from zenml.constants import METADATA_EXPERIMENT_TRACKER_URL
from zenml.experiment_trackers.base_experiment_tracker import (
    BaseExperimentTracker,
)
from zenml.integrations.mlflow import mlflow_utils
from zenml.integrations.mlflow.flavors.mlflow_experiment_tracker_flavor import (
    MLFlowExperimentTrackerConfig,
    MLFlowExperimentTrackerSettings,
    is_databricks_tracking_uri,
    is_remote_mlflow_tracking_uri,
)
from zenml.logger import get_logger
from zenml.metadata.metadata_types import Uri
from zenml.stack import StackValidator

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo
    from zenml.metadata.metadata_types import MetadataType

logger = get_logger(__name__)

MLFLOW_TRACKING_URI = "MLFLOW_TRACKING_URI"
MLFLOW_TRACKING_USERNAME = "MLFLOW_TRACKING_USERNAME"
MLFLOW_TRACKING_PASSWORD = "MLFLOW_TRACKING_PASSWORD"
MLFLOW_TRACKING_TOKEN = "MLFLOW_TRACKING_TOKEN"
MLFLOW_TRACKING_INSECURE_TLS = "MLFLOW_TRACKING_INSECURE_TLS"
MLFLOW_BACKEND_STORE_URI = "_MLFLOW_SERVER_FILE_STORE"

DATABRICKS_HOST = "DATABRICKS_HOST"
DATABRICKS_USERNAME = "DATABRICKS_USERNAME"
DATABRICKS_PASSWORD = "DATABRICKS_PASSWORD"
DATABRICKS_TOKEN = "DATABRICKS_TOKEN"
DATABRICKS_UNITY_CATALOG = "databricks-uc"


class MLFlowExperimentTracker(BaseExperimentTracker):
    """Track experiments using MLflow."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the experiment tracker and validate the tracking uri.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._ensure_valid_tracking_uri()

    def _ensure_valid_tracking_uri(self) -> None:
        """Ensures that the tracking uri is a valid mlflow tracking uri.

        Raises:
            ValueError: If the tracking uri is not valid.
        """
        tracking_uri = self.config.tracking_uri
        if tracking_uri:
            valid_schemes = DATABASE_ENGINES + ["http", "https", "file"]
            if not any(
                tracking_uri.startswith(scheme) for scheme in valid_schemes
            ) and not is_databricks_tracking_uri(tracking_uri):
                raise ValueError(
                    f"MLflow tracking uri does not start with one of the valid "
                    f"schemes {valid_schemes} or its value is not set to "
                    f"'databricks'. See "
                    f"https://www.mlflow.org/docs/latest/tracking.html#where-runs-are-recorded "
                    f"for more information."
                )

    @property
    def config(self) -> MLFlowExperimentTrackerConfig:
        """Returns the `MLFlowExperimentTrackerConfig` config.

        Returns:
            The configuration.
        """
        return cast(MLFlowExperimentTrackerConfig, self._config)

    @property
    def local_path(self) -> Optional[str]:
        """Path to the local directory where the MLflow artifacts are stored.

        Returns:
            None if configured with a remote tracking URI, otherwise the
            path to the local MLflow artifact store directory.
        """
        tracking_uri = self.get_tracking_uri()
        if is_remote_mlflow_tracking_uri(tracking_uri):
            return None
        else:
            assert tracking_uri.startswith("file:")
            return tracking_uri[5:]

    @property
    def validator(self) -> Optional["StackValidator"]:
        """Checks the stack has a `LocalArtifactStore` if no tracking uri was specified.

        Returns:
            An optional `StackValidator`.
        """
        if self.config.tracking_uri:
            # user specified a tracking uri, do nothing
            return None
        else:
            # try to fall back to a tracking uri inside the zenml artifact
            # store. this only works in case of a local artifact store, so we
            # make sure to prevent stack with other artifact stores for now
            return StackValidator(
                custom_validation_function=lambda stack: (
                    isinstance(stack.artifact_store, LocalArtifactStore),
                    "MLflow experiment tracker without a specified tracking "
                    "uri only works with a local artifact store.",
                )
            )

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Mlflow experiment tracker.

        Returns:
            The settings class.
        """
        return MLFlowExperimentTrackerSettings

    @staticmethod
    def _local_mlflow_backend() -> str:
        """Gets the local MLflow backend inside the ZenML artifact repository directory.

        Returns:
            The MLflow tracking URI for the local MLflow backend.
        """
        client = Client()
        artifact_store = client.active_stack.artifact_store
        local_mlflow_tracking_uri = os.path.join(artifact_store.path, "mlruns")
        if not os.path.exists(local_mlflow_tracking_uri):
            os.makedirs(local_mlflow_tracking_uri)
        return "file:" + local_mlflow_tracking_uri

    def get_tracking_uri(self, as_plain_text: bool = True) -> str:
        """Returns the configured tracking URI or a local fallback.

        Args:
            as_plain_text: Whether to return the tracking URI as plain text.

        Returns:
            The tracking URI.
        """
        if as_plain_text:
            tracking_uri = self.config.tracking_uri
        else:
            tracking_uri = self.config.model_dump()["tracking_uri"]
        return tracking_uri or self._local_mlflow_backend()

    def prepare_step_run(self, info: "StepRunInfo") -> None:
        """Sets the MLflow tracking uri and credentials.

        Args:
            info: Info about the step that will be executed.
        """
        self.configure_mlflow()
        settings = cast(
            MLFlowExperimentTrackerSettings,
            self.get_settings(info),
        )

        experiment_name = settings.experiment_name or info.pipeline.name
        experiment = self._set_active_experiment(experiment_name)
        run_id = self.get_run_id(
            experiment_name=experiment_name, run_name=info.run_name
        )

        # Validate that the run exists before attempting to resume it
        if run_id:
            try:
                run = mlflow.get_run(run_id)
                if run.info.lifecycle_stage == LifecycleStage.DELETED:
                    logger.warning(
                        f"Run with id {run_id} is in 'deleted' state. "
                        "Creating a new run instead."
                    )
                    run_id = None
            except MlflowException as e:
                # Run doesn't exist on the MLflow server, create a new one
                logger.warning(
                    f"Run with id {run_id} not found in MLflow tracking server. "
                    f"Creating a new run instead. Error: {e}"
                )
                run_id = None

        tags = settings.tags.copy()
        tags.update(self._get_internal_tags())

        mlflow.start_run(
            run_id=run_id,
            run_name=info.run_name,
            experiment_id=experiment.experiment_id,
            tags=tags,
        )

        if settings.nested:
            mlflow.start_run(
                run_name=info.pipeline_step_name, nested=True, tags=tags
            )

    def get_step_run_metadata(
        self, info: "StepRunInfo"
    ) -> Dict[str, "MetadataType"]:
        """Get component- and step-specific metadata after a step ran.

        Args:
            info: Info about the step that was executed.

        Returns:
            A dictionary of metadata.
        """
        metadata: Dict[str, Any] = {
            METADATA_EXPERIMENT_TRACKER_URL: Uri(
                self.get_tracking_uri(as_plain_text=False)
            ),
        }
        if run := mlflow.active_run():
            metadata["mlflow_run_id"] = run.info.run_id
            metadata["mlflow_experiment_id"] = run.info.experiment_id

        return metadata

    def disable_autologging(self) -> None:
        """Disables MLflow autologging for all supported frameworks."""
        frameworks = [
            "tensorflow",
            "xgboost",
            "lightgbm",
            "statsmodels",
            "spark",
            "sklearn",
            "fastai",
            "pytorch",
        ]

        failed_frameworks = []

        for framework in frameworks:
            try:
                # Correctly prefix the module name with 'mlflow.'
                module_name = f"mlflow.{framework}"
                # Dynamically import the module corresponding to the framework
                module = importlib.import_module(module_name)
                # Call the autolog function with disable=True
                module.autolog(disable=True)
            except ImportError as e:
                # only log on mlflow relevant errors
                if "mlflow" in e.msg.lower():
                    failed_frameworks.append(framework)
            except Exception:
                failed_frameworks.append(framework)

        if len(failed_frameworks) > 0:
            logger.warning(
                f"Failed to disable MLflow autologging for the following frameworks: "
                f"{failed_frameworks}."
            )

    def cleanup_step_run(
        self,
        info: "StepRunInfo",
        step_failed: bool,
    ) -> None:
        """Stops active MLflow runs and resets the MLflow tracking uri.

        Args:
            info: Info about the step that was executed.
            step_failed: Whether the step failed or not.
        """
        status = "FAILED" if step_failed else "FINISHED"
        self.disable_autologging()
        mlflow_utils.stop_zenml_mlflow_runs(status)
        mlflow.set_tracking_uri("")

    def configure_mlflow(self) -> None:
        """Configures the MLflow tracking URI and any additional credentials."""
        tracking_uri = self.get_tracking_uri()
        mlflow.set_tracking_uri(tracking_uri)

        if is_databricks_tracking_uri(tracking_uri):
            if self.config.databricks_host:
                os.environ[DATABRICKS_HOST] = self.config.databricks_host
            if self.config.tracking_username:
                os.environ[DATABRICKS_USERNAME] = self.config.tracking_username
            if self.config.tracking_password:
                os.environ[DATABRICKS_PASSWORD] = self.config.tracking_password
            if self.config.tracking_token:
                os.environ[DATABRICKS_TOKEN] = self.config.tracking_token
            if self.config.enable_unity_catalog:
                mlflow.set_registry_uri(DATABRICKS_UNITY_CATALOG)
        else:
            os.environ[MLFLOW_TRACKING_URI] = tracking_uri
            if self.config.tracking_username:
                os.environ[MLFLOW_TRACKING_USERNAME] = (
                    self.config.tracking_username
                )
            if self.config.tracking_password:
                os.environ[MLFLOW_TRACKING_PASSWORD] = (
                    self.config.tracking_password
                )
            if self.config.tracking_token:
                os.environ[MLFLOW_TRACKING_TOKEN] = self.config.tracking_token

        os.environ[MLFLOW_TRACKING_INSECURE_TLS] = (
            "true" if self.config.tracking_insecure_tls else "false"
        )

    def get_run_id(self, experiment_name: str, run_name: str) -> Optional[str]:
        """Gets the if of a run with the given name and experiment.

        Args:
            experiment_name: Name of the experiment in which to search for the
                run.
            run_name: Name of the run to search.

        Returns:
            The id of the run if it exists.
        """
        self.configure_mlflow()
        experiment_name = self._adjust_experiment_name(experiment_name)

        runs = mlflow.search_runs(
            experiment_names=[experiment_name],
            filter_string=f'tags.mlflow.runName = "{run_name}"',
            run_view_type=3,
            output_format="list",
        )
        if not runs:
            return None

        run: Run = runs[0]
        if mlflow_utils.is_zenml_run(run):
            return cast(str, run.info.run_id)
        else:
            return None

    def _set_active_experiment(self, experiment_name: str) -> Experiment:
        """Sets the active MLflow experiment.

        If no experiment with this name exists, it is created and then
        activated.

        Args:
            experiment_name: Name of the experiment to activate.

        Raises:
            RuntimeError: If the experiment creation or activation failed.

        Returns:
            The experiment.
        """
        experiment_name = self._adjust_experiment_name(experiment_name)

        mlflow.set_experiment(experiment_name=experiment_name)
        experiment = mlflow.get_experiment_by_name(experiment_name)
        if not experiment:
            raise RuntimeError("Failed to set active mlflow experiment.")
        return experiment

    def _adjust_experiment_name(self, experiment_name: str) -> str:
        """Prepends a slash to the experiment name if using Databricks.

        Databricks requires the experiment name to be an absolute path within
        the Databricks workspace.

        Args:
            experiment_name: The experiment name.

        Returns:
            The potentially adjusted experiment name.
        """
        tracking_uri = self.get_tracking_uri()

        if (
            tracking_uri
            and is_databricks_tracking_uri(tracking_uri)
            and not experiment_name.startswith("/")
        ):
            return f"/{experiment_name}"
        else:
            return experiment_name

    @staticmethod
    def _get_internal_tags() -> Dict[str, Any]:
        """Gets ZenML internal tags for MLflow runs.

        Returns:
            Internal tags.
        """
        return {mlflow_utils.ZENML_TAG_KEY: zenml.__version__}



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for pipeline runs."""

from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.constants import (
    API,
    DISABLE_HEARTBEAT,
    LOGS,
    LOGS_MAX_ENTRIES_PER_REQUEST,
    LOGS_RUNNER_SOURCE,
    PIPELINE_CONFIGURATION,
    REFRESH,
    RUNS,
    STATUS,
    STEPS,
    STOP,
    VERSION_1,
)
from zenml.enums import ExecutionStatus
from zenml.logger import get_logger
from zenml.models import (
    Page,
    PipelineRunDAG,
    PipelineRunFilter,
    PipelineRunRequest,
    PipelineRunResponse,
    PipelineRunUpdate,
    StepRunFilter,
    StepRunResponse,
)
from zenml.utils import run_utils
from zenml.utils.logging_utils import (
    LogEntry,
    fetch_logs,
    search_logs_by_source,
)
from zenml.zen_server.auth import (
    AuthContext,
    authorize,
)
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_get_or_create_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import (
    dehydrate_response_model,
    verify_permission_for_model,
)
from zenml.zen_server.routers.projects_endpoints import workspace_router
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    server_config,
    workload_manager,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + RUNS,
    tags=["runs"],
    responses={401: error_response, 403: error_response},
)


logger = get_logger(__name__)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.post(
    "/{project_name_or_id}" + RUNS,
    responses={401: error_response, 409: error_response, 422: error_response},
    deprecated=True,
    tags=["runs"],
)
@async_fastapi_endpoint_wrapper
def get_or_create_pipeline_run(
    pipeline_run: PipelineRunRequest,
    project_name_or_id: Optional[Union[str, UUID]] = None,
    _: AuthContext = Security(authorize),
) -> Tuple[PipelineRunResponse, bool]:
    """Get or create a pipeline run.

    Args:
        pipeline_run: Pipeline run to create.
        project_name_or_id: Optional name or ID of the project.

    Returns:
        The pipeline run and a boolean indicating whether the run was created
        or not.
    """
    if project_name_or_id:
        project = zen_store().get_project(project_name_or_id)
        pipeline_run.project = project.id

    if pipeline_run.original_run_id:
        # Verify that the user has permissions to read the original run.
        # so they cannot gain unauthorized access to some step configurations
        # of the original run.
        original_run = zen_store().get_run(
            pipeline_run.original_run_id, hydrate=False
        )
        verify_permission_for_model(model=original_run, action=Action.READ)

    return verify_permissions_and_get_or_create_entity(
        request_model=pipeline_run,
        get_or_create_method=zen_store().get_or_create_run,
    )


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.get(
    "/{project_name_or_id}" + RUNS,
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
    tags=["runs"],
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def list_runs(
    runs_filter_model: PipelineRunFilter = Depends(
        make_dependable(PipelineRunFilter)
    ),
    project_name_or_id: Optional[Union[str, UUID]] = None,
    hydrate: bool = False,
    include_full_metadata: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[PipelineRunResponse]:
    """Get pipeline runs according to query filters.

    Args:
        runs_filter_model: Filter model used for pagination, sorting, filtering.
        project_name_or_id: Optional name or ID of the project.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.
        include_full_metadata: Flag deciding whether to include the
            full metadata in the response.

    Returns:
        The pipeline runs according to query filters.
    """
    if project_name_or_id:
        runs_filter_model.project = project_name_or_id

    return verify_permissions_and_list_entities(
        filter_model=runs_filter_model,
        resource_type=ResourceType.PIPELINE_RUN,
        list_method=zen_store().list_runs,
        hydrate=hydrate,
        include_full_metadata=include_full_metadata,
    )


@router.get(
    "/{run_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def get_run(
    run_id: UUID,
    hydrate: bool = True,
    refresh_status: bool = False,
    include_python_packages: bool = False,
    include_full_metadata: bool = False,
    _: AuthContext = Security(authorize),
) -> PipelineRunResponse:
    """Get a specific pipeline run using its ID.

    Args:
        run_id: ID of the pipeline run to get.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.
        refresh_status: Flag deciding whether we should try to refresh
            the status of the pipeline run using its orchestrator.
        include_python_packages: Flag deciding whether to include the
            Python packages in the response.
        include_full_metadata: Flag deciding whether to include the
            full metadata in the response.

    Returns:
        The pipeline run.
    """
    store = zen_store()
    run = verify_permissions_and_get_entity(
        id=run_id,
        get_method=store.get_run,
        hydrate=hydrate,
        include_python_packages=include_python_packages,
        include_full_metadata=include_full_metadata,
    )

    if refresh_status:
        try:
            logger.warning(
                "DEPRECATED: The ability to refresh the status a run through "
                "the GET `/runs/{run_id}` endpoint is deprecated and will be "
                "removed in a future version. Please use the POST "
                "`/runs/{run_id}/refresh` endpoint instead."
            )
            run = run_utils.refresh_run_status(run=run, zen_store=store)

        except Exception as e:
            logger.warning(
                "An error occurred while refreshing the status of the "
                f"pipeline run: {e}"
            )
    return run


@router.put(
    "/{run_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def update_run(
    run_id: UUID,
    run_model: PipelineRunUpdate,
    _: AuthContext = Security(authorize),
) -> PipelineRunResponse:
    """Updates a run.

    Args:
        run_id: ID of the run.
        run_model: Run model to use for the update.

    Returns:
        The updated run model.
    """
    return verify_permissions_and_update_entity(
        id=run_id,
        update_model=run_model,
        get_method=zen_store().get_run,
        update_method=zen_store().update_run,
    )


@router.delete(
    "/{run_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_run(
    run_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a run.

    Args:
        run_id: ID of the run.
    """
    verify_permissions_and_delete_entity(
        id=run_id,
        get_method=zen_store().get_run,
        delete_method=zen_store().delete_run,
    )


@router.get(
    "/{run_id}" + STEPS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def get_run_steps(
    run_id: UUID,
    step_run_filter_model: StepRunFilter = Depends(
        make_dependable(StepRunFilter)
    ),
    _: AuthContext = Security(authorize),
) -> Page[StepRunResponse]:
    """Get all steps for a given pipeline run.

    Args:
        run_id: ID of the pipeline run.
        step_run_filter_model: Filter model used for pagination, sorting,
            filtering

    Returns:
        The steps for a given pipeline run.
    """
    verify_permissions_and_get_entity(
        id=run_id, get_method=zen_store().get_run, hydrate=False
    )
    step_run_filter_model.pipeline_run_id = run_id
    return zen_store().list_run_steps(step_run_filter_model)


@router.get(
    "/{run_id}" + PIPELINE_CONFIGURATION,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_pipeline_configuration(
    run_id: UUID,
    _: AuthContext = Security(authorize),
) -> Dict[str, Any]:
    """Get the pipeline configuration of a specific pipeline run using its ID.

    Args:
        run_id: ID of the pipeline run to get.

    Returns:
        The pipeline configuration of the pipeline run.
    """
    run = verify_permissions_and_get_entity(
        id=run_id, get_method=zen_store().get_run, hydrate=True
    )
    return run.config.model_dump()


@router.get(
    "/{run_id}" + STATUS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_run_status(
    run_id: UUID,
    _: AuthContext = Security(authorize),
) -> ExecutionStatus:
    """Get the status of a specific pipeline run.

    Args:
        run_id: ID of the pipeline run for which to get the status.

    Returns:
        The status of the pipeline run.
    """
    run = verify_permissions_and_get_entity(
        id=run_id, get_method=zen_store().get_run, hydrate=False
    )
    return run.status


@router.get(
    "/{run_id}/dag",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_run_dag(
    run_id: UUID,
    _: AuthContext = Security(authorize),
) -> PipelineRunDAG:
    """Get the DAG of a specific pipeline run.

    Args:
        run_id: ID of the pipeline run for which to get the DAG.

    Returns:
        The DAG of the pipeline run.
    """
    # TODO: Maybe avoid calling get_run twice?
    verify_permissions_and_get_entity(
        id=run_id,
        get_method=zen_store().get_run,
        hydrate=False,
    )
    return zen_store().get_pipeline_run_dag(run_id)


@router.post(
    "/{run_id}" + REFRESH,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def refresh_run_status(
    run_id: UUID,
    include_steps: bool = False,
    _: AuthContext = Security(authorize),
) -> None:
    """Refreshes the status of a specific pipeline run.

    Args:
        run_id: ID of the pipeline run to refresh.
        include_steps: Flag deciding whether we should also refresh
            the status of individual steps.
    """
    store = zen_store()
    run = verify_permissions_and_get_entity(
        id=run_id,
        get_method=store.get_run,
        hydrate=True,
    )
    run_utils.refresh_run_status(
        run=run, include_step_updates=include_steps, zen_store=store
    )


@router.post(
    "/{run_id}" + STOP,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def stop_run(
    run_id: UUID,
    graceful: bool = False,
    _: AuthContext = Security(authorize),
) -> None:
    """Stops a specific pipeline run.

    Args:
        run_id: ID of the pipeline run to stop.
        graceful: If True, allows for graceful shutdown where possible.
            If False, forces immediate termination. Default is False.
    """
    run = zen_store().get_run(run_id, hydrate=True)
    verify_permission_for_model(run, action=Action.READ)
    verify_permission_for_model(run, action=Action.UPDATE)
    dehydrate_response_model(run)
    run_utils.stop_run(run=run, graceful=graceful)


@router.get(
    "/{run_id}" + LOGS,
    responses={
        400: error_response,
        401: error_response,
        404: error_response,
        422: error_response,
    },
)
@async_fastapi_endpoint_wrapper
def run_logs(
    run_id: UUID,
    source: str,
    _: AuthContext = Security(authorize),
) -> List[LogEntry]:
    """Get log entries for efficient pagination.

    This endpoint returns the log entries.

    Args:
        run_id: ID of the pipeline run.
        source: Required source to get logs for.

    Returns:
        List of log entries.

    Raises:
        KeyError: If no logs are found for the specified source.
    """
    store = zen_store()

    run = verify_permissions_and_get_entity(
        id=run_id, get_method=store.get_run, hydrate=True
    )

    # Handle runner logs from workload manager
    if run.snapshot and source == LOGS_RUNNER_SOURCE:
        snapshot = run.snapshot

        # The runner source is explicitly added to the log collection for runs
        # with version >0.94.0. For legacy runs, we didn't specify that
        # explicitly and instead rely on source snapshot or trigger existence.
        runner_logs_response = search_logs_by_source(
            run.log_collection or [], source
        )
        is_legacy_run_with_runner_logs = (
            snapshot.template_id or snapshot.source_snapshot_id or run.trigger
        )

        if (
            runner_logs_response or is_legacy_run_with_runner_logs
        ) and server_config().workload_manager_enabled:
            from zenml.log_stores.artifact.artifact_log_store import (
                parse_log_entry,
            )

            if run.trigger:
                # Trigger invocation
                workload_id = run.id
            elif run.source_snapshot:
                # Manual snapshot run. When we resume a run that was initially
                # triggered by a snapshot, we'll only show the runner logs
                # of the initial snapshot run. Not sure how to adjust this in
                # the future, maybe multiple runner log sources, that point to
                # specific workload IDs?
                workload_id = snapshot.id
            else:
                # Resume/Retry run from server
                workload_id = run.id

            workload_logs = workload_manager().get_logs(
                workload_id=workload_id
            )

            log_entries = []
            for line in workload_logs.split("\n"):
                if log_record := parse_log_entry(line):
                    log_entries.append(log_record)

                if len(log_entries) >= LOGS_MAX_ENTRIES_PER_REQUEST:
                    break

            return log_entries

    if run.log_collection:
        if logs_response := search_logs_by_source(run.log_collection, source):
            return fetch_logs(
                logs=logs_response,
                zen_store=store,
                limit=LOGS_MAX_ENTRIES_PER_REQUEST,
            )

    raise KeyError(f"No logs found for source '{source}' in run {run_id}")


@router.put(
    "/{run_id}" + DISABLE_HEARTBEAT,
    responses={
        400: error_response,
        401: error_response,
        404: error_response,
        422: error_response,
    },
)
@async_fastapi_endpoint_wrapper
def disable_run_heartbeat(
    run_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Disables heartbeats for a run.

    Args:
        run_id: ID of the run.
    """
    run = verify_permissions_and_get_entity(
        id=run_id, get_method=zen_store().get_run, hydrate=False
    )

    verify_permission_for_model(run, action=Action.READ)
    verify_permission_for_model(run, action=Action.UPDATE)

    zen_store().disable_run_heartbeat(run_id=run_id)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes deployer flavor."""

from typing import TYPE_CHECKING, Dict, List, Optional, Type

from pydantic import Field, PositiveInt, field_validator, model_validator

from zenml.constants import KUBERNETES_CLUSTER_RESOURCE_TYPE
from zenml.deployers.base_deployer import (
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.enums import KubernetesServiceType
from zenml.integrations.kubernetes import (
    KUBERNETES_DEPLOYER_FLAVOR,
)
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.logger import get_logger
from zenml.models import ServiceConnectorRequirements

if TYPE_CHECKING:
    from zenml.integrations.kubernetes.deployers import KubernetesDeployer

logger = get_logger(__name__)


class KubernetesDeployerSettings(BaseDeployerSettings):
    """Settings for the Kubernetes deployer.

    Follows a progressive complexity model:
    - Essential Settings: Core configuration for Deployment + Service (80% of users)
    - Additional Resources: Add ANY K8s resource (Ingress, HPA, NetworkPolicy, etc.)
    - Custom Templates: Override deployment.yaml.j2 or service.yaml.j2 for advanced control

    Essential Settings:
        namespace: Kubernetes namespace for deployments.
        service_type: How to expose service (LoadBalancer, NodePort, ClusterIP).
        service_port: Port to expose on the service.
        image_pull_policy: When to pull images (Always, IfNotPresent, Never).
        labels: Custom labels to apply to all resources.
        command: Override container command/entrypoint.
        args: Override container args.
        service_account_name: ServiceAccount for pods (RBAC).
        image_pull_secrets: Secrets for private registries.

    Advanced Settings:
        pod_settings: Advanced pod configuration (volumes, affinity, tolerations, etc.).
        additional_resources: Paths to YAML files with additional K8s resources.
        custom_templates_dir: Path to custom Jinja2 templates for Deployment/Service.
          Use this to customize health probes, service annotations, etc.
    """

    namespace: str = Field(
        "zenml-deployments",
        description="Default Kubernetes namespace for deployments. "
        "Can be overridden per-deployment using the `namespace` setting. "
        "Defaults to 'zenml-deployments'.",
    )

    service_type: KubernetesServiceType = Field(
        default=KubernetesServiceType.LOAD_BALANCER,
        description=(
            "Type of Kubernetes Service: LoadBalancer, NodePort, or ClusterIP. "
            "LoadBalancer is recommended for production (requires cloud provider support). "
            "For production with custom domains and TLS, consider using ClusterIP with an Ingress."
        ),
    )

    service_port: PositiveInt = Field(
        default=8000,
        description="Port to expose on the service.",
    )

    image_pull_policy: str = Field(
        default="IfNotPresent",
        description="Kubernetes image pull policy: Always, IfNotPresent, or Never.",
    )

    labels: Dict[str, str] = Field(
        default_factory=dict,
        description="Additional labels to apply to all Kubernetes resources.",
    )

    annotations: Dict[str, str] = Field(
        default_factory=dict,
        description="Annotations to add to Pod resources. "
        "Example: {'prometheus.io/scrape': 'true', 'prometheus.io/port': '8000'}",
    )

    # Container configuration
    command: Optional[List[str]] = Field(
        default=None,
        description="Override container command (entrypoint). "
        "If not set, uses the image's default or ZenML's deployment server.",
    )

    args: Optional[List[str]] = Field(
        default=None,
        description="Override container args. "
        "If not set, uses defaults appropriate for the deployment.",
    )

    service_account_name: Optional[str] = Field(
        default=None,
        description="Kubernetes service account for the deployment pods. "
        "If not set, uses the default service account in the namespace.",
    )

    image_pull_secrets: List[str] = Field(
        default_factory=list,
        description="Names of Kubernetes secrets for pulling private images. "
        "Example: ['my-registry-secret', 'dockerhub-secret']",
    )

    # ========================================================================
    # Health Probes (necessary for production robustness)
    # ========================================================================

    readiness_probe_path: str = Field(
        default="/api/health",
        description="HTTP path for readiness probe.",
    )

    readiness_probe_initial_delay: PositiveInt = Field(
        default=10,
        description="Initial delay in seconds before starting readiness probe.",
    )

    readiness_probe_period: PositiveInt = Field(
        default=10,
        description="How often (in seconds) to perform readiness probe.",
    )

    readiness_probe_timeout: PositiveInt = Field(
        default=5,
        description="Timeout in seconds for readiness probe.",
    )

    readiness_probe_failure_threshold: PositiveInt = Field(
        default=3,
        description="Number of failures before marking pod as not ready.",
    )

    liveness_probe_path: str = Field(
        default="/api/health",
        description="HTTP path for liveness probe.",
    )

    liveness_probe_initial_delay: PositiveInt = Field(
        default=30,
        description="Initial delay in seconds before starting liveness probe.",
    )

    liveness_probe_period: PositiveInt = Field(
        default=10,
        description="How often (in seconds) to perform liveness probe.",
    )

    liveness_probe_timeout: PositiveInt = Field(
        default=5,
        description="Timeout in seconds for liveness probe.",
    )

    liveness_probe_failure_threshold: PositiveInt = Field(
        default=3,
        description="Number of failures before restarting pod.",
    )

    # ========================================================================
    # Advanced Pod Configuration
    # ========================================================================

    pod_settings: Optional[KubernetesPodSettings] = Field(
        default=None,
        description="Advanced pod configuration: volumes, affinity, tolerations, node selectors, etc. "
        "Use this for advanced scheduling and resource requirements.",
    )

    strict_additional_resources: bool = Field(
        default=True,
        description="If True (default), fail the deployment if any additional resource fails to apply. "
        "If False, log warnings for failed resources but continue with core deployment. "
        "Recommended: True for production to ensure all critical resources (HPA, NetworkPolicy, etc.) are applied.",
    )

    additional_resources: List[str] = Field(
        default_factory=list,
        description="Paths to YAML files with additional Kubernetes resources (Ingress, HPA, etc.). "
        "Supports Jinja2 templating with deployment context variables like {{namespace}}, {{service_name}}, "
        "and {{labels}}. Use `---` to separate multiple resources in one file. "
        "See docs for available template variables and examples.",
    )

    atomic_provision: bool = Field(
        default=False,
        description="If True, attempt to rollback (delete) all provisioned resources if deployment fails. "
        "Mimics Helm's --atomic flag behavior. This is best-effort cleanup, not a true transaction. "
        "If False (default), partial resources remain on failure for inspection and manual cleanup. "
        "The resource inventory is always tracked regardless of this setting.",
    )

    # ========================================================================
    # Template and Development Settings
    # ========================================================================

    custom_deployment_template_file: Optional[str] = Field(
        default=None,
        description="Path to custom Jinja2 template file for Kubernetes Deployment. "
        "Overrides the built-in deployment.yaml.j2 template. "
        "Can be a local path or remote URL. "
        "Example: '~/.zenml/k8s-templates/my-deployment.yaml.j2'",
    )

    custom_service_template_file: Optional[str] = Field(
        default=None,
        description="Path to custom Jinja2 template file for Kubernetes Service. "
        "Overrides the built-in service.yaml.j2 template. "
        "Can be a local path or remote URL. "
        "Example: '~/.zenml/k8s-templates/my-service.yaml.j2'",
    )

    # ========================================================================
    # Internal/Backward Compatibility Properties
    # ========================================================================

    wait_for_load_balancer_timeout: int = Field(
        default=150,
        ge=0,
        description="Timeout in seconds for LoadBalancer IP assignment. "
        "Set to 0 to skip waiting. Only applies to LoadBalancer service type.",
    )

    deployment_ready_check_interval: PositiveInt = Field(
        default=2,
        description="Interval in seconds between deployment readiness checks.",
    )

    @field_validator("readiness_probe_path", "liveness_probe_path")
    @classmethod
    def validate_probe_path(cls, v: str) -> str:
        """Validate that probe paths start with /.

        Args:
            v: The probe path value.

        Returns:
            The validated probe path.

        Raises:
            ValueError: If the path doesn't start with /.
        """
        if not v.startswith("/"):
            raise ValueError(f"Probe path must start with '/': {v}")
        return v

    @model_validator(mode="after")
    def validate_interdependent_settings(self) -> "KubernetesDeployerSettings":
        """Validate settings that depend on each other.

        Returns:
            The validated settings.
        """
        # Warn if wait_for_load_balancer_timeout is set for non-LoadBalancer services
        if (
            self.service_type != "LoadBalancer"
            and self.wait_for_load_balancer_timeout > 0
        ):
            logger.warning(
                f"wait_for_load_balancer_timeout={self.wait_for_load_balancer_timeout} "
                f"is ignored for service_type={self.service_type}. "
                f"This setting only applies to LoadBalancer service type."
            )

        return self


class KubernetesDeployerConfig(BaseDeployerConfig, KubernetesDeployerSettings):
    """Configuration for the Kubernetes deployer.

    This config combines deployer-specific settings with Kubernetes
    component configuration (context, namespace, in-cluster mode).

    Attributes:
        incluster: If `True`, the deployer will run inside the same cluster in which it itself is running. This requires the client to run in a Kubernetes pod itself. If set, the `kubernetes_context` config option is ignored. If the stack component is linked to a Kubernetes service connector, this field is ignored.
        kubernetes_context: Name of a Kubernetes context to run deployments in. If the stack component is linked to a Kubernetes service connector, this field is ignored. Otherwise, it is mandatory.
        kubernetes_namespace: Default Kubernetes namespace for deployments. Can be overridden per-deployment using the `namespace` setting. Defaults to 'zenml-deployments'.
        local: If `True`, the deployer will assume it is connected to a local kubernetes cluster and will perform additional validations.
    """

    incluster: bool = Field(
        False,
        description="If `True`, the deployer will run inside the "
        "same cluster in which it itself is running. This requires the client "
        "to run in a Kubernetes pod itself. If set, the `kubernetes_context` "
        "config option is ignored. If the stack component is linked to a "
        "Kubernetes service connector, this field is ignored.",
    )

    kubernetes_context: Optional[str] = Field(
        None,
        description="Name of a Kubernetes context to run deployments in. "
        "If the stack component is linked to a Kubernetes service connector, "
        "this field is ignored. Otherwise, it is mandatory.",
    )

    local: bool = Field(
        False,
        description="If `True`, the deployer will assume it is connected to a "
        "local kubernetes cluster and will perform additional validations.",
    )

    skip_local_validations: bool = Field(
        False,
        description="If `True`, the local validations will be skipped.",
    )

    @property
    def is_local(self) -> bool:
        """Checks if this is a local Kubernetes cluster.

        Returns:
            True if using a local Kubernetes cluster, False otherwise.
        """
        if self.local:
            return True
        if self.kubernetes_context:
            local_context_indicators = [
                "k3d-",
                "kind-",
                "minikube",
                "docker-desktop",
                "colima",
                "rancher-desktop",
            ]
            context_lower = self.kubernetes_context.lower()
            return any(
                indicator in context_lower
                for indicator in local_context_indicators
            )
        return False

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return not self.is_local


class KubernetesDeployerFlavor(BaseDeployerFlavor):
    """Flavor for the Kubernetes deployer."""

    @property
    def name(self) -> str:
        """The name of the flavor.

        Returns:
            The flavor name.
        """
        return KUBERNETES_DEPLOYER_FLAVOR

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector requirements for the Kubernetes deployer.

        Returns:
            Service connector requirements.
        """
        return ServiceConnectorRequirements(
            resource_type=KUBERNETES_CLUSTER_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A URL to docs about this flavor.

        Returns:
            The documentation URL.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A URL to SDK docs about this flavor.

        Returns:
            The SDK documentation URL.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """The logo URL for the flavor.

        Returns:
            The logo URL.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/kubernetes.png"

    @property
    def config_class(self) -> Type[KubernetesDeployerConfig]:
        """Returns `KubernetesDeployerConfig` config class.

        Returns:
            The config class.
        """
        return KubernetesDeployerConfig

    @property
    def implementation_class(self) -> Type["KubernetesDeployer"]:
        """Returns the implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.kubernetes.deployers import (
            KubernetesDeployer,
        )

        return KubernetesDeployer
